(function() {
	'use strict';

	angular.module('ipconfigure.orchid.directive').directive('eventsChart', ['$window', eventsChart]);

	function eventsChart($window) {
		function hourMinuteFormatter(d) {
			if (d.getHours() === 0 && d.getMinutes() === 0) {
				return d3.time.format('%m/%d')(d);
			}
			else {
				return d3.time.format('%H:%M')(d);
			}
		}

		return {
			restrict: 'E',
			scope: {
				events: '=',
				from: '=',
				to: '=',
				onPointClick: '&'
			},
			replace: true,
			template: '<div><svg></svg></div>',
			link: function(scope, element) {
				var initializeChecksum = 0,
					xScale,
					yScale,
					xAxis,
					yAxis,
					svg,
					margin = {
						top: 5,
						right: 30,
						bottom: 15,
						left: 30
					},
					points = [],
					yMaxValue,
					yMinValue,
					barWidth = 10,
					barTotalMargin = 5;

				function initializeChart() {
					// wait for scope.from, scope.to, and scope.events before doing anything
					initializeChecksum += 1;

					// once we have all three values we're expecting via $watch, initialize the svg
					if (initializeChecksum === 3) {
						xScale = d3.time.scale()
							.domain([scope.from, scope.to])
							.range([0, element[0].clientWidth - 100]);

						yScale = d3.scale.linear()
							.domain([0, yMaxValue])
							.range([element[0].clientHeight - margin.top - margin.bottom, 0]);

						xAxis = d3.svg.axis()
							.scale(xScale)
							.orient('bottom')
							.tickSize(1)
							.tickFormat(function(d) {
								return hourMinuteFormatter(d);
							});

						yAxis = d3.svg.axis()
							.scale(yScale)
							.orient('left')
							.ticks(Math.min(8, yMaxValue - yMinValue + 1))
							.tickSize(-element[0].clientWidth);

						svg = d3.select(element[0]).select('svg')
							.attr('width', '100%')
							.style('height', '100%')
							.append('g')
							.attr('transform', 'translate(' + margin.left + ',' + margin.top + ')')
							.on('dblclick', function() {
								var events;

								if (d3.event.target.tagName.toLowerCase() === 'rect' && (d3.event.target.getAttribute('class') === null || d3.event.target.getAttribute('class').indexOf('filler') === -1)) {
									events = d3.select(d3.event.target).data()[0].events;

									if (events.length > 1 && events[events.length - 1].when - events[0].when > 60000) {
										scope.from = new Date(events[0].when);
										scope.to = new Date(events[events.length - 1].when);

										scope.from.setMilliseconds(0);
										scope.from.setSeconds(0);
										scope.to.setMilliseconds(0);
										if (scope.to.getSeconds() > 0) {
											scope.to.setSeconds(0);
											scope.to.setMinutes(scope.to.getMinutes() + 1);
										}

										scope.$apply();
									}
									else {
										alert('Cannot zoom in any further.');
									}
								}
							})
							.on('click', function() {
								// if the user has clicked a rect, raise the onPointClick event with all the related events
								if (d3.event.target.tagName.toLowerCase() === 'rect' && (d3.event.target.getAttribute('class') === null || d3.event.target.getAttribute('class').indexOf('filler') === -1)) {
									// highlight the one they clicked
									svg.select('g.points').selectAll('rect').style('fill', '');
									d3.select(d3.event.target).style('fill', 'red');

									scope.onPointClick({ events: d3.select(d3.event.target).data()[0].events });
									scope.$apply();
								}
								else {
									// clear selection
									svg.select('g.points').selectAll('rect').style('fill', '');

									scope.onPointClick({ events: [] });
									scope.$apply();
								}
							});

						svg.append('rect')
							.attr('class', 'filler')
							.attr('width', '100%')
							.attr('height', '100%');

						svg.append('g')
							.attr('class', 'x axis')
							.attr('transform', 'translate(0,' + (element[0].clientHeight - margin.top - margin.bottom) + ')')
							.call(xAxis);

						svg.append('g')
							.attr('class', 'y axis')
							.call(yAxis);

						svg.append('g')
							.attr('class', 'points');
					}

					return initializeChecksum >= 3;
				}

				function adjustDates() {
					if (angular.isDefined(scope.from) === true && angular.isDefined(scope.to) === true && initializeChart() === true) {
						xScale.domain([scope.from, scope.to]);
						render();
					}
				}

				function render(skipAnimation) {
					var duration = skipAnimation ? 0 : 250,
						data = svg.select('g.points')
							.selectAll('rect')
							.data(points, function(d) { return d.x.getTime(); });

					data.exit()
						.attr('class', 'removed')
						.transition()
						.duration(duration)
						.attr('y', function() { return element[0].clientHeight - margin.top - margin.bottom; })
						.attr('height', 0)
						.remove();

					data.enter()
						.append('rect')
						.attr('width', barWidth)
						.attr('x', function(d) { return xScale(d.x) - barWidth / 2; })
						.attr('y', function() { return element[0].clientHeight - margin.top - margin.bottom; })
						.attr('height', 0);

					// this isn't right is it? it's re-selecting the entered ones as well? i forget
					svg.select('g.points')
						.selectAll('rect:not(.removed)')
						.transition()
						.delay(duration)
						.duration(duration)
						.attr('x', function(d) {
							return xScale(d.x) - barWidth / 2;
						})
						.attr('y', function(d) {
							d.h = element[0].clientHeight - margin.top - margin.bottom - yScale(d.y);

							return yScale(d.y) - Math.abs((Math.max(10, d.h) - d.h));
						})
						.attr('width', barWidth)
						.attr('height', function(d) { return Math.max(10, d.h); });

					svg.selectAll('g .y.axis')
						.transition()
						.duration(duration)
						.call(yAxis);

					svg.selectAll('g .x.axis')
						.transition()
						.duration(duration)
						.call(xAxis);
				}

				$window.addEventListener('resize', function() {
					xScale.range([0, element[0].clientWidth - 100]);
					yAxis.tickSize(-element[0].clientWidth);
					render();
				});

				scope.$watch('to', adjustDates);
				scope.$watch('from', adjustDates);

				scope.$watch('events', function() {
					var i,
						len,
						j,
						aggregatedPoints = {},
						roundedWhen,
						barSpanTime;

					// aggregate event data into buckets
					points = [];

					if (scope.events.length > 0) {
						// define buckets based on barWidth starting at 0
						i = scope.from.getTime() + (xScale.invert(barWidth / 2 + 1).getTime() - xScale.invert(0).getTime());
						len = scope.to.getTime();
						barSpanTime = xScale.invert(barWidth + barTotalMargin) - xScale.invert(0);

						// fill buckets
						while (i < len) {
							for (j = 0; j < scope.events.length; j += 1) {
								if (scope.events[j].when >= i && scope.events[j].when <= i + barSpanTime) {
									if (!aggregatedPoints[i]) {
										aggregatedPoints[i] = [];
									}

									aggregatedPoints[i].push(scope.events[j]);
								}
							}

							i += barSpanTime;
						}

						// populate points based on filled buckets
						for (roundedWhen in aggregatedPoints) {
							if (aggregatedPoints.hasOwnProperty(roundedWhen)) {
								points.push({
									x: new Date(parseInt(roundedWhen, 10)),
									y: aggregatedPoints[roundedWhen].length,
									events: aggregatedPoints[roundedWhen]
								});
							}
						}
					}

					if (initializeChart() === true) {
						// build min and max
						if (points.length > 0) {
							yMinValue = points.reduce(function(a, b) { return a < b.y ? a : b.y; }, 1);
							yMaxValue = points.reduce(function(a, b) { return a > b.y ? a : b.y; }, 0);
						}
						else {
							yMinValue = 1;
							yMaxValue = 1;
						}

						yScale.domain([0, yMaxValue]);
						yAxis.ticks(Math.min(8, yMaxValue - yMinValue + 1));

						render();
					}
				});
			}
		};
	}
}());
