(function($console) {
	'use strict';

	window.ipConfigure = window.ipConfigure || {};
	ipConfigure.components = ipConfigure.components || {};

	/**
	 * Timeline class, injects a timeline the element indicated by the id parameter.
	 * Exposes events and allows control of the timeline.
	 *
	 * @param {string} id
	 * @param {object} options
	 *
	 * @namespace ipConfigure.components.timeline
	 */
	ipConfigure.components.timeline = function(id, options) {
		var that = this,
			_container = d3.select('#' + id),
			_width = _container.node().clientWidth,
			_svg = _container.append('svg'),
			_mode = '', // live, playback, or empty string?
			_xOffset = _container.node().getBoundingClientRect().left,
			_currentTime,
			_timeIndicator = {
				dragging: false,
				width: 120,
				indicatorLabelFormat: d3.time.format('%b %-d  %-H:%M:%S')
			},
			_liveIndicator = {
				width: 50
			},
			_fineBar = {
				dragging: false,
				height: 63,
				yOffset: 17,
				xOffset: 0,
				tickTimeFormat: function(d) { return _formatters.fine(d); },
				liveBar: {},
				halfTimeSteps: [
					60000 / 2, // 1 minute
					60000 * 2, // 4 minutes
					60000 * 4, // 8 minutes
					60000 * 16, // 32 minutes
					60000 * 60, // 2 hours
					60000 * 60 * 6, // 12 hours
					60000 * 60 * 12, // 1 day
					60000 * 60 * 24, // 2 days
					60000 * 60 * 24 * 3, // 6 days
					60000 * 60 * 24 * 5], // 10 days
				liveHalfTime: 60000 * 4
			},
			_coarseBar = {
				configurations: [],
				configurationIndex: 0,
				height: 25,
				yOffset: 0,
				leftRightAdjustment: 0,
				segmentWidth: 100,
				scale: d3.time.scale(),
				timeWindow: {},
				timeWindowHandleWidth: 20,
				timeWindowTriangleWidth: 15,
				timeWindowMinimumWidth: 17,
				liveBar: {}
			},
			_motionEventData = [],
			_motionEventsAnchored = false,
			_defaultOptions = {
				debug: false,
				interactive: true,
				draggableIndicator: true,
				onTimeChange: function(time, source) {
					// where time is time in milliseconds
					// and source is either 'via-fine' or 'via-needle' or 'via-live' or 'via-coarse'
					// this event also signifies that the user is no longer ripping
					_log('onTimeChange event not subscribed to: ', time, source);
				},
				onCoarseTimeChange: function(startTime, stopTime) {
					_log('onCoarseTimeChange event not subscribed to: ', startTime, stopTime);
				},
				onEndReached: function() {
					_log('onEndReached');
				},
				onScaleAdjust: function(configurationIndex) {
					_log('onScaleAdjust event not subscribed to: ', configurationIndex);
				},
				onRip: function(time, source) {
					// where time is time in milliseconds
					// and source is either 'via-fine' or 'via-needle'
					_log('onRip event not subscribed to: ', time, source);
				}
			},
			_options = ipConfigure.core.mergeOptions(_defaultOptions, options),
			_milliseconds = {
				perDay: 1000 * 60 * 60 * 24,
				perHour: 1000 * 60 * 60,
				perMinute: 1000 * 60
			},
			_formatters = {
				fineChecksum: {},
				fine: function(d) {
					var hour = d.getHours(),
						minute = d.getMinutes(),
						day = d.getDate(),
						key;

					if (minute === 0 && hour === 0) {
						key = d.getMonth().toString() + day.toString() + hour.toString() + minute.toString();

						if (_formatters.fineChecksum[key] !== true) {
							_formatters.fineChecksum[key] = true;

							return d3.time.format('%b %-d')(d) + _d3TimeOrdinalSuffix(day);
						}
					}

					return d3.time.format('%-H:%M:%S')(d);
				},
				month: d3.time.format('%B'),
				monthDay: function(d) {
					var day = d.getDate();

					if (day === 1) {
						return _formatters.month(d);
					}
					else {
						return d3.time.format('%b %-d')(d) + _d3TimeOrdinalSuffix(day);
					}
				},
				day: function(d) {
					var day = d.getDate();
					return (day === 1) ? _formatters.month(d) : day;
				},
				minuteMinute: d3.time.format('%-H:%M'),
				minute: function(d) {
					var hour = d.getHours(),
						minute = d.getMinutes();

					if (minute === 0 && hour === 0) {
						return _formatters.monthDay(d);
					}
					else {
						return _formatters.minuteMinute(d).toLowerCase();
					}
				}
			};

		function _log() {
			if (options.debug === true) {
				$console.log.apply($console, arguments);
			}
		}

		// fine events

		function _onFineBarDragStart(d) {
			// if the fine bar is interactive and it's the left mouse button (1) or a touch event (0)
			if (_options.interactive === true && (d3.event.sourceEvent.which === 0 || d3.event.sourceEvent.which === 1)) {
				d3.event.sourceEvent.stopPropagation();

				if (typeof d3.event.sourceEvent.clientX === 'undefined') {
					// touch fix
					if (window.TouchEvent && d3.event.sourceEvent instanceof window.TouchEvent) {
						d3.event.sourceEvent.clientX = d3.event.sourceEvent.changedTouches[0].clientX;
						d3.event.sourceEvent.clientY = d3.event.sourceEvent.changedTouches[0].clientY;
					}
				}

				_xOffset = _container.node().getBoundingClientRect().left;

				// firefox hack
				if (typeof d3.event.sourceEvent.offsetX === 'undefined') {
					d3.event.sourceEvent.offsetX = d3.event.sourceEvent.layerX - _xOffset;
				}

				d.startX = d3.event.sourceEvent.clientX - _xOffset;
				d.startOffsetX = d3.event.sourceEvent.offsetX;

				d.startTime = _fineBar.start;
				d.stopTime = _fineBar.stop;
			}
		}

		function _onFineBarDrag(d) {
			var msPerPixel,
				coarseBarData,
				minimumStart,
				maximumStart,
				calculatedStart,
				calculatedStop,
				calculatedCurrent,
				calculatedScale = d3.time.scale().range(_fineBar.scale.range()),
				now = Date.now();

			if (_options.interactive === true && (d3.event.sourceEvent.which === 0 || d3.event.sourceEvent.which === 1)) {
				if (typeof d3.event.sourceEvent.clientX === 'undefined') {
					// touch fix
					if (window.TouchEvent && d3.event.sourceEvent instanceof window.TouchEvent) {
						d3.event.sourceEvent.clientX = d3.event.sourceEvent.changedTouches[0].clientX;
						d3.event.sourceEvent.clientY = d3.event.sourceEvent.changedTouches[0].clientY;
					}
				}

				_fineBar.dragging = true;

				// determine what the length of a pixel is
				msPerPixel = _fineBar.scale.invert(1) - _fineBar.scale.invert(0);

				// determine the new start/stop time
				calculatedStart = d.startTime + msPerPixel * (d.startX - d3.event.sourceEvent.clientX + _xOffset);

				if (_coarseBar.created === true && _mode !== 'live') {
					coarseBarData = _coarseBar.timeWindow.data()[0];
					minimumStart = _coarseBar.scale.invert(_coarseBar.timeWindowTriangleWidth).getTime();
					maximumStart = _coarseBar.scale.invert(_width - coarseBarData.width - _coarseBar.timeWindowTriangleWidth).getTime();
					calculatedStart = Math.max(calculatedStart, minimumStart);
					calculatedStart = Math.min(calculatedStart, maximumStart);
				}

				calculatedStop = calculatedStart + (d.stopTime - d.startTime);

				// determine where the indicator is pointing with the new calculatedStart/Stop values
				calculatedScale.domain([new Date(calculatedStart), new Date(calculatedStop)]);

				if (_mode === 'live') {
					calculatedCurrent = calculatedScale.invert(_timeIndicator.container.data()[0].x).getTime();
				}
				else {
					// if in playback, restrict this to 50%
					calculatedCurrent = calculatedScale.invert(Math.round(_width * 0.50)).getTime();
				}

				// prevent the user from dragging into the future
				if (_mode === 'live' && calculatedCurrent > now) {
					calculatedStart -= calculatedCurrent - now;
					calculatedStop -= calculatedCurrent - now;
					calculatedCurrent = now;
				}

				_currentTime = calculatedCurrent;
				_fineBar.start = calculatedStart;
				_fineBar.stop = calculatedStop;

				// re-render the fine bar
				_fineRenderTimeline();
				_positionFineLiveIndicator(); // redundant, except for unlinked players who don't still call setTime when ripping =/

				if (_mode === 'live') {
					// raise event rip event
					_options.onRip(_currentTime, {
						source: 'via-fine',
						timelineWidth: _calculateCurrentWidth(),
						fineStart: _fineBar.start,
						fineStop: _fineBar.stop,
						fineScale: _fineBar.scale
					});

					_positionFineIndicator(_calculateCurrentWidth() / 2);

					// update the fine indicator label
					_updateFineIndicatorLabelFromTime(calculatedCurrent);
				}

				_setTimeWindowPosition(_fineBar.start, _fineBar.stop, false);
			}
		}

		function _onFineBarDragEnd(d) {
			var indicatedTime,
				nowX;

			if (_options.interactive === true && (d3.event.sourceEvent.which === 0 || d3.event.sourceEvent.which === 1)) {
				// since dragging is sometimes set to true, despite no drag actually happening (chrome/mouse issues), we're
				// checking to see if the x changed... if not, no drag occurred
				if (_fineBar.dragging === true && d.startX !== d3.event.sourceEvent.clientX - _xOffset) {
					_fineBar.dragging = false;

					indicatedTime = _fineBar.scale.invert(_timeIndicator.container.data()[0].x).getTime();

					if (_mode === 'live') {
						// user has dragged the fine bar while the indicated was snapped to the middle
						_options.onTimeChange(indicatedTime, 'via-fine');
					}
					else if (_mode === 'playback' && (indicatedTime < _fineBar.start || indicatedTime > _fineBar.stop)) {
						// the user has dragged the current time out of view, yank it back to the left
						_options.onTimeChange(_fineBar.start, 'via-fine');
					}
				}
				else {
					// always need to reset this, in case it was set above
					_fineBar.dragging = false;

					// since they didn't drag, assume they clicked and figure out if they clicked on the live blue bar
					nowX = _fineBar.scale(Date.now());

					// use d.startOffsetX since d.startX is relative to the entire viewport
					if (d.startOffsetX >= nowX) {
						_options.onTimeChange(Date.now(), 'via-live');
					}
					else {
						_options.onTimeChange(_fineBar.scale.invert(d.startX - _fineBar.xOffset).getTime(), 'via-fine-click');
					}
				}
			}
		}

		function _onFineBarMouseWheel() {
			var wheelEvent = d3.event.sourceEvent,
				step,
				originalHalfTime,
				i,
				len;

			if (wheelEvent === null) {
				// https://github.com/mbostock/d3/issues/2411
				return;
			}

			if (_options.interactive === true) {
				if (_mode === 'live' && (wheelEvent.type === 'wheel' || wheelEvent.type === 'mousewheel')) {
					// prevent this event from trigger on elements behind the timeline
					d3.event.sourceEvent.stopPropagation();

					step = wheelEvent.wheelDelta ? (wheelEvent.wheelDelta > 0 ? -1 : 1) : wheelEvent.detail ? (wheelEvent.detail > 0 ? 1 : -1) : (wheelEvent.deltaY > 0 ? 1 : -1);
					originalHalfTime = _fineBar.liveHalfTime;

					// we have configured half time values that we step through... we need to calculate the new liveHalfTime value by stepping through our
					// configured values, since if the user adjusted the time window, we may end up with an odd liveHalfTime

					if (step === 1) {
						// mouse wheel up, zoom out for a larger fine timeline
						for (i = 0, len = _fineBar.halfTimeSteps.length; i < len; i++) {
							if (_fineBar.halfTimeSteps[i] > _fineBar.liveHalfTime) {
								_fineBar.liveHalfTime = _fineBar.halfTimeSteps[i];
								break;
							}
						}
					}
					else {
						// mouse wheel down, zoom in for a larger fine timeline
						for (i = _fineBar.halfTimeSteps.length - 1, len = -1; i > len; i--) {
							if (_fineBar.halfTimeSteps[i] < _fineBar.liveHalfTime) {
								_fineBar.liveHalfTime = _fineBar.halfTimeSteps[i];
								break;
							}
						}
					}

					if (_fineBar.liveHalfTime !== originalHalfTime) {
						_fineBar.start += originalHalfTime;
						_fineBar.stop -= originalHalfTime;
						_fineBar.start -= _fineBar.liveHalfTime;
						_fineBar.stop += _fineBar.liveHalfTime;

						_fineRenderTimeline(true);
					}
				}
			}
		}

		function _onFineIndicatorDragStart() {
			if (_options.interactive === true) {
				d3.event.sourceEvent.stopPropagation();
			}
		}

		function _onFineIndicatorDrag(d) {
			var now,
				maxX,
				minX,
				calculatedX,
				calculatedDate;

			if (_options.interactive === true && _options.draggableIndicator === true) {
				now = new Date();
				maxX = _fineBar.xOffset > 0 ? _calculateCurrentWidth() : _calculateCurrentWidth() - 9;
				minX = _fineBar.xOffset > 0 ? 0 : 25; // where 25 is the width of a drag handle (so there's a smooth transition to coarse)
				calculatedX = Math.min(maxX, Math.max(minX, d3.event.x));
				calculatedDate = _fineBar.scale.invert(calculatedX);

				_timeIndicator.dragging = true;

				if (calculatedDate < now) {
					d.x = calculatedX;
				}
				else {
					d.x = _fineBar.scale(now);
				}

				// raise event rip event
				_options.onRip(((calculatedDate < now) ? calculatedDate : now).getTime(), {
					source: 'via-needle',
					timelineWidth: _calculateCurrentWidth(),
					fineStart: _fineBar.start,
					fineStop: _fineBar.stop,
					fineScale: _fineBar.scale
				});

				_positionFineIndicator(d.x);
				_updateFineIndicatorLabelFromPoint(d.x);
			}
		}

		function _onFineIndicatorDragEnd(d) {
			if (_options.interactive === true && _options.draggableIndicator === true && _timeIndicator.dragging === true) {
				_timeIndicator.dragging = false;

				_options.onTimeChange(_fineBar.scale.invert(d.x).getTime(), 'via-needle');
			}
		}

		function _onFineRightHandleDragStart() {
		}

		function _onFineRightHandleDrag(d) {
			var minDate,
				minX;

			if (_options.interactive === true) {
				d.dragged = true;
				d.x = d3.event.x;

				if (d.x > _width) {
					d.x = _width;

					_fineBar.rightHandleGrabber
						.attr('transform', 'translate(' + (_width) + ')')
						.data([
							{ x: _width }
						]);

					_fineBar.rightHandle
						.attr('width', _width);

					//  slide right
					if (!_fineBar.resizeTimer) {
						_fineBar.resizeTimer = setInterval(function() {
							var centerTime = _fineBar.scale.invert(_width / 2).getTime(),
								diff,
								coarseBarEnd = _coarseBar.scale.invert(_width - _coarseBar.timeWindowTriangleWidth).getTime();

							// + one half of a percent of the current time window
							_fineBar.stop += parseInt((_fineBar.stop - _fineBar.start) * 0.005, 10);

							if (centerTime > Date.now()) {
								// calculate difference
								diff = centerTime - Date.now();

								_fineBar.stop -= diff;
							}

							// prevent the user from dragging the time window off the coarse bar
							if (_fineBar.stop > coarseBarEnd) {
								// calculate difference
								diff = _fineBar.stop - coarseBarEnd;

								_fineBar.stop -= diff;
							}

							_fineRenderTimeline();
							_setTimeWindowPosition(new Date(_fineBar.start), new Date(_fineBar.stop), false);
						}, 1000 / 30);
					}
				}
				else {
					if (_fineBar.resizeTimer) {
						clearInterval(_fineBar.resizeTimer);
						delete _fineBar.resizeTimer;
					}

					minDate = _fineBar.scale.invert(25);
					minDate.setSeconds(minDate.getSeconds() + 10);
					minDate.setMilliseconds(minDate.getMilliseconds() + 1);
					minX = _fineBar.scale(minDate);

					if (d.x <= minX) {
						d.x = minX;
					}

					// color the lines on the right handle to white if past now
					if (_fineBar.scale(Date.now()) > d.x - 12) {
						// white
						_fineBar.rightHandleGrabber.selectAll('line').style('stroke', 'white');
					}
					else {
						// not white
						_fineBar.rightHandleGrabber.selectAll('line').style('stroke', '#c8e3f1');
					}

					_setTimeWindowPosition(new Date(_fineBar.start), _fineBar.scale.invert(d.x), false);

					_fineBar.rightHandleGrabber
						.attr('transform', 'translate(' + (d.x) + ')')
						.data([
							{ x: d.x }
						]);

					_fineBar.rightHandle
						.attr('width', _width - d.x);
				}
			}
		}

		function _onFineRightHandleDragEnd(d) {
			var newStopDate;

			if (_options.interactive === true && d.dragged === true) {
				if (_fineBar.resizeTimer) {
					clearInterval(_fineBar.resizeTimer);
					delete _fineBar.resizeTimer;
				}

				// determine the new stop date
				newStopDate = _fineBar.scale.invert(d.x - _fineBar.xOffset);

				// make adjustments to the timeline
				_fineBar.stop = newStopDate.getTime();

				// and render it
				_fineRenderTimeline(true, function() {
					_positionFineIndicatorAfterAdjustment('via-fine-resize');
				});

				// reset the drag bar position
				d.x = _calculateCurrentWidth() + _fineBar.xOffset;

				_fineBar.rightHandleGrabber
					.data([
						{ x: d.x }
					])
					.transition()
					.attr('transform', 'translate(' + (d.x) + ')');

				_fineBar.rightHandle
					.transition()
					.attr('width', _fineBar.xOffset);
			}
		}

		function _onFineLeftHandleDragStart() {
		}

		function _onFineLeftHandleDrag(d) {
			var minDate,
				minX;

			if (_options.interactive === true) {
				d.dragged = true;

				d.x = d3.event.x;

				if (d.x < 0) {
					_fineBar.leftHandle.attr('width', 0);

					// slide left
					if (!_fineBar.resizeTimer) {
						_fineBar.resizeTimer = setInterval(function() {
							// - one half of a percent of the current time window
							_fineBar.start -= parseInt((_fineBar.stop - _fineBar.start) * 0.005, 10);

							_fineRenderTimeline();
							_setTimeWindowPosition(new Date(_fineBar.start), new Date(_fineBar.stop), false);
						}, 1000 / 30);
					}
				}
				else {
					if (_fineBar.resizeTimer) {
						clearInterval(_fineBar.resizeTimer);
						delete _fineBar.resizeTimer;
					}

					minDate = _fineBar.scale.invert(_width - 25);
					minDate.setSeconds(minDate.getSeconds() - 10);
					minDate.setMilliseconds(minDate.getMilliseconds() + 1);
					minX = _fineBar.scale(minDate);

					if (d.x > minX) {
						d.x = minX;
					}

					_setTimeWindowPosition(_fineBar.scale.invert(d.x), new Date(_fineBar.stop), false);

					_fineBar.leftHandle.attr('width', d.x);
					_fineBar.leftHandlGrabberLineGroup.attr('transform', 'translate(' + (d.x - 25 + 9) + ')');
				}
			}
		}

		function _onFineLeftHandleDragEnd(d) {
			var newStartDate;

			if (_options.interactive === true && d.dragged === true) {
				if (_fineBar.resizeTimer) {
					clearInterval(_fineBar.resizeTimer);
					delete _fineBar.resizeTimer;
				}

				// determine the new start date
				newStartDate = _fineBar.scale.invert(d.x - 25);

				// make adjustments to the timeline
				_fineBar.start = newStartDate.getTime();

				// and render it
				_fineRenderTimeline(true, function() {
					_positionFineIndicatorAfterAdjustment('via-fine-resize');
				});

				// reset the drag bar position
				d.x = _fineBar.xOffset;
				_fineBar.leftHandle
					.transition()
					.attr('width', _fineBar.xOffset);
				_fineBar.leftHandlGrabberLineGroup
					.transition()
					.attr('transform', 'translate(9)');
			}
		}

		// coarse events

		function _onCoarseBarDragStart() {
			d3.event.sourceEvent.stopPropagation();

			_coarseBar.dragging = true;
		}

		function _onCoarseBarDrag() {
			var d = _coarseBar.segmentContainer.data()[0],
				forceRedraw = false,
				currentConfiguration = _coarseBar.configurations[_coarseBar.configurationIndex],
				timeWindowData,
				xAdjustment,
				dontGoPast = (_coarseBar.timeWindow.data()[0].x + (_coarseBar.timeWindow.data()[0].width * 0.50)),
				dx,
				step;

			xAdjustment = d3.event.x + (_coarseBar.leftRightAdjustment * _coarseBar.segmentWidth);
			dx = xAdjustment - d.x;

			_log(_coarseBar.liveBar.container.data()[0].x, ' + ', dx, ' < ', dontGoPast, '??', xAdjustment);
			d.x = xAdjustment;

			// if the user has dragged the bar to the left, we need to adjust some things
			if (d.x > 0) {
				// figure out how many segments they moved over and subtract the right bits
				step = Math.ceil(d.x / _coarseBar.segmentWidth);
				d.x -= step * _coarseBar.segmentWidth;
				_coarseBar.startTime -= step * currentConfiguration.milliseconds;
				_coarseBar.leftRightAdjustment -= step;

				forceRedraw = true;
			}
			else if (d.x < -(_coarseBar.segmentWidth * 2)) {
				// figure out how many segments they moved over and add the right bits
				step = Math.floor(Math.abs(d.x) / (_coarseBar.segmentWidth * 2));
				d.x += step * _coarseBar.segmentWidth;
				_coarseBar.startTime += step * currentConfiguration.milliseconds;
				_coarseBar.leftRightAdjustment += step;

				forceRedraw = true;
			}

			_coarseBar.segmentContainer
				.attr('transform', function(segmentContainerData) {
					return 'translate(' + segmentContainerData.x + ', 0)';
				});

			if (forceRedraw === true) {
				_coarseRenderTimeline();
			}

			_updateCoarseScaleRange(d.x);

			// re-render the fine bar
			timeWindowData = _coarseBar.timeWindow.data()[0];
			_fineBar.start = _coarseBar.scale.invert(timeWindowData.x).getTime();
			_fineBar.stop = _coarseBar.scale.invert(timeWindowData.x + timeWindowData.width).getTime();
			_fineRenderTimeline();
		}

		function _onCoarseBarDragEnd() {
			_coarseBar.dragging = false;
			_coarseBar.leftRightAdjustment = 0;

			_positionFineIndicatorAfterAdjustment('via-coarse');
		}

		function _onCoarseBarClick() {
			if (_options.interactive === true && d3.event.defaultPrevented === false) {
				var d = _coarseBar.segmentContainer.data()[0],
					offset = d.x,
					atSegment = Math.floor((d3.event.pageX - _xOffset - offset) / _coarseBar.segmentWidth),
					timeWindowData = _coarseBar.timeWindow.data()[0];

				timeWindowData.x = (atSegment * _coarseBar.segmentWidth) + (_coarseBar.segmentWidth + offset) - _coarseBar.segmentWidth - (_coarseBar.segmentWidth / 2);
				timeWindowData.width = (_coarseBar.segmentWidth * 2);
				timeWindowData.x = Math.max(timeWindowData.x, 11);

				// reposition the time window
				_positionTimeWindow(timeWindowData, true);

				// re-render the fine bar
				_fineBar.start = _coarseBar.scale.invert(timeWindowData.x).getTime();
				_fineBar.stop = _coarseBar.scale.invert(timeWindowData.x + timeWindowData.width).getTime();
				_fineRenderTimeline(true);

				// raise onTimeChange if necessary
				_positionFineIndicatorAfterAdjustment('via-coarse');

				_coarseBar.timeWindow.dragging = false;
			}
		}

		function _onCoarseLiveBarClick() {
			if (_options.interactive === true) {
				_options.onTimeChange(Date.now(), 'via-live');
			}
		}

		// time window events

		function _onTimeWindowDragStart(d) {
			d3.event.sourceEvent.stopPropagation();

			_coarseBar.timeWindow.dragging = true;

			if (typeof d3.event.sourceEvent.clientX === 'undefined') {
				// touch fix
				if (window.TouchEvent && d3.event.sourceEvent instanceof window.TouchEvent) {
					d3.event.sourceEvent.clientX = d3.event.sourceEvent.changedTouches[0].clientX;
					d3.event.sourceEvent.clientY = d3.event.sourceEvent.changedTouches[0].clientY;
				}
			}

			// keep track of where the user started dragging and how long the bar was
			d.initialOffset = d3.event.sourceEvent.clientX - _xOffset - d.x;
			d.initialWidth = d.width;

			// keep track of the element that was dragged so that we can correctly position things later on
			d.element = d3.event.sourceEvent.target;
		}

		function _onTimeWindowDrag(d) {
			var className = d.element.className.baseVal;

			if (className === 'time-window-left-handle') {
				_onTimeWindowDragLeftHandle(d);
			}
			else if (className === 'time-window-center') {
				_onTimeWindowDragCenter(d);
			}
			else if (className === 'time-window-right-handle') {
				_onTimeWindowDragRightHandle(d);
			}

			// re-render the fine bar
			_fineBar.start = _coarseBar.scale.invert(d.x).getTime();
			_fineBar.stop = _coarseBar.scale.invert(d.x + d.width).getTime();

			// reposition the time window
			_positionTimeWindow(d, false);

			// render the timeline
			_fineRenderTimeline();
		}

		function _onTimeWindowDragLeftHandle(d) {
			var minimumX,
				calculatedX,
				calculatedW;

			// determine the minimum position of the time window in case the user ripped their mouse past the point
			minimumX = d.x + d.width - _coarseBar.timeWindowMinimumWidth;

			// determine where the left edge of the time window should go
			calculatedX = Math.max(_coarseBar.timeWindowTriangleWidth, d3.event.x - d.initialOffset);

			// determine how long should the center block be based on the new left edge
			calculatedW = d.width + d.x - calculatedX;

			if (calculatedW < 0) {
				// if the user has dragged past the right edge, then we need to switch which element we're dragging
				d.element = _coarseBar.timeWindowRightHandle[0][0];
				d.initialOffset = _coarseBar.timeWindowMinimumWidth;
				d.initialWidth = _coarseBar.timeWindowMinimumWidth;
				d.width = _coarseBar.timeWindowMinimumWidth;
				d.x = minimumX;
			}
			else if (calculatedW > _coarseBar.timeWindowMinimumWidth) {
				// the user is just resizing the left edge, so use the calculated values
				d.x = calculatedX;
				d.width = calculatedW;
			}
			else {
				// we've reached the minimum width, so flatten the time window
				d.width = _coarseBar.timeWindowMinimumWidth;
				d.x = minimumX;
			}
		}

		function _onTimeWindowDragCenter(d) {
			var x;

			// determine where the left edge of the time window should go, while constraining the bounds, not letting it blow out of its container
			x = Math.min(_width - d.width - _coarseBar.timeWindowTriangleWidth, Math.max(_coarseBar.timeWindowTriangleWidth, d3.event.x - d.initialOffset));

			d.x = x;
		}

		function _onTimeWindowDragRightHandle(d) {
			// determine the inner width of the time window
			var offset = (d.initialOffset - d.initialWidth),
				innerWidth = (d3.event.x - d.x) - offset,
				width;

			if (innerWidth < -offset) {
				// if the user has dragged past the left edge, then we need to switch which element we're dragging
				d.element = _coarseBar.timeWindowLeftHandle[0][0];
				d.initialOffset = 0;
				d.initialWidth = 0;
				d.width = _coarseBar.timeWindowMinimumWidth;
			}
			else {
				// constrain the width (not letting it go negative, and not letting it blow out of its container
				width = Math.min(_width - d.x - _coarseBar.timeWindowTriangleWidth, Math.max(_coarseBar.timeWindowMinimumWidth, innerWidth));

				d.width = width;
			}
		}

		function _onTimeWindowDragEnd() {
			_coarseBar.timeWindow.dragging = false;

			_positionFineIndicatorAfterAdjustment('via-coarse');
		}

		// creation functions

		function _createSvgContainer() {
			var defs = _svg.append('defs'),
				filter;

			// width is necessary for FireFox
			_svg.attr('width', '100%');

			// we initialize with this height, but it may grow later if the coarse bar is shown
			_container.style('height', _fineBar.height + 'px');

			// set the width/height
			_svg
				.attr('class', 'timeline');

			// create defs

			// indicatorNeedleDropShadow
			filter = defs.append('filter')
				.attr('id', 'indicatorNeedleDropShadow')
				.attr('x', '-10%')
				.attr('y', '-10%')
				.attr('width', '175%')
				.attr('height', '175%');
			filter.append('feOffset')
				.attr('result', 'offOut')
				.attr('in', 'SourceAlpha')
				.attr('dx', 1)
				.attr('dy', 0);
			filter.append('feGaussianBlur')
				.attr('result', 'blurOut')
				.attr('in', 'offOut')
				.attr('stdDeviation', 1);
			filter.append('feBlend')
				.attr('in', 'SourceGraphic')
				.attr('in2', 'blurOut')
				.attr('mode', 'normal');

			// indicatorShapeDropShadow
			filter = defs.append('filter')
				.attr('id', 'indicatorShapeDropShadow')
				.attr('x', '-10%')
				.attr('y', '-10%')
				.attr('width', '140%')
				.attr('height', '140%');
			filter.append('feOffset')
				.attr('result', 'offOut')
				.attr('in', 'SourceAlpha')
				.attr('dx', 2)
				.attr('dy', 0);
			filter.append('feGaussianBlur')
				.attr('result', 'blurOut')
				.attr('in', 'offOut')
				.attr('stdDeviation', 2);
			filter.append('feBlend')
				.attr('in', 'SourceGraphic')
				.attr('in2', 'blurOut')
				.attr('mode', 'normal');

			// indicatorLabelDropShadow
			filter = defs.append('filter')
				.attr('id', 'indicatorLabelDropShadow')
				.attr('x', '-10%')
				.attr('y', '-10%')
				.attr('width', '140%')
				.attr('height', '140%');
			filter.append('feOffset')
				.attr('result', 'offOut')
				.attr('in', 'SourceAlpha')
				.attr('dx', 0)
				.attr('dy', 0);
			filter.append('feGaussianBlur')
				.attr('result', 'blurOut')
				.attr('in', 'offOut')
				.attr('stdDeviation', 3);
			filter.append('feBlend')
				.attr('in', 'SourceGraphic')
				.attr('in2', 'blurOut')
				.attr('mode', 'normal');

			// create the fine bar container
			_fineBar.container = _svg.append('g')
				.attr('class', 'fine');

			// create the coarse bar container
			_coarseBar.container = _svg.append('g')
				.attr('class', 'coarse');

			// create the time indicator container
			_timeIndicator.container = _svg.append('g')
				.attr('class', 'fine-indicator');

			// create the live indicator container
			_liveIndicator.container = _svg.append('g')
				.attr('class', 'live-indicator clickable');

		}

		function _createFineBar() {
			var ticks;

			if (_fineBar.created !== true) {
				// initialize the scale
				_fineBar.scale = d3.time.scale()
					.range([0, _width]);

				ticks = _calculateTicksForWidth();

				// initialize the axis
				_fineBar.axis = d3.svg.axis()
					.scale(_fineBar.scale)
					.ticks(ticks)
					.tickSize(14, 0)
					.tickPadding(5)
					.tickFormat(_fineBar.tickTimeFormat);

				// create the backdrop (css will apply color)
				_fineBar.container.append('rect')
					.attr('class', 'fine-background')
					.attr('width', '100%')
					.attr('height', (_fineBar.height - _fineBar.yOffset))
					.attr('transform', 'translate(0, ' + _fineBar.yOffset + ')');

				// create a container for the axis
				_fineBar.axisGroup = _fineBar.container.append('g')
					.attr('class', 'x axis');

				// create a container for the fake ticks
				_fineBar.axisMinorGroup = _fineBar.container.append('g')
					.attr('class', 'x axis minor');

				_fineBar.axisMotionGroup = _fineBar.container.append('g')
					.attr('class', 'x axis motion');

				// create the fine live indicator
				_fineBar.liveBar.container = _fineBar.container.append('g')
					.attr('class', 'live');

				// create a draggable rect that appears in front of the fine bar to make it draggable
				_fineBar.container.append('rect')
					.attr('class', 'fine-foreground')
					.attr('width', '100%')
					.attr('height', (_fineBar.height - _fineBar.yOffset))
					.attr('transform', 'translate(0, ' + _fineBar.yOffset + ')')
					.call(d3.behavior.zoom()
						.on('zoom', _onFineBarMouseWheel))
					.call(d3.behavior.drag()
						.origin(Object)
						.on('dragstart', _onFineBarDragStart)
						.on('drag', _onFineBarDrag)
						.on('dragend', _onFineBarDragEnd))
					.data([
						{}
					]);

				// create the left/right resizing containerspro
				_fineBar.sizerGroup = _fineBar.container.append('g');

				// create the LIVE indicator
				_fineBar.liveBar.rect = _fineBar.liveBar.container
					.append('rect')
					.attr('height', _fineBar.height - _fineBar.yOffset);

				// create the top drop shadow - http://docs.webplatform.org/wiki/svg/tutorials/smarter_svg_graphics
				_fineBar.container.append('pattern')
					.attr('id', 'dropShadowTilePattern')
					.attr('x', '0')
					.attr('y', '0')
					.attr('width', '11')
					.attr('height', '80')
					.attr('patternContentUnits', 'userSpaceOnUse')
					.attr('patternUnits', 'userSpaceOnUse')
					.append('image')
					.attr('id', 'dropShadowTileRect')
					.attr('xlink:href', 'img/fade.png')
					.attr('x', '0')
					.attr('y', '0')
					.attr('width', '63px')
					.attr('height', '8px');

				_fineBar.container.append('rect')
					.attr('x', '0')
					.attr('y', 0)
					.attr('width', '100%')
					.attr('height', '5')
					.attr('fill', 'url(#dropShadowTilePattern)')
					.attr('transform', 'translate(0, ' + _fineBar.yOffset + ')');

				// create the time indicator
				_createFineIndicator();

				// create the live indicator
				_createLiveIndicator();

				// don't let this ever happen again
				_fineBar.created = true;

				// initialize this so we can write cleaner code
				_fineBar.dragging = false;
			}
		}

		function _createFineIndicator() {
			if (_timeIndicator.created !== true) {
				_timeIndicator.created = true;

				_timeIndicator.pointerGroup = _timeIndicator.container.append('g')
					.attr('class', 'pointer');

				// tick line (drop shadows on lines apparently don't work in IE/FF, so I'm using a rect)
				_timeIndicator.pointerGroup.append('rect')
					.attr('x', -1)
					.attr('y', 0)
					.attr('width', 2)
					.attr('height', _fineBar.height);

				// funky shape
				if (_options.draggableIndicator === true) {
					_timeIndicator.container.attr('class', 'fine-indicator draggable');

					_timeIndicator.pointerGroup.append('polygon')
						.attr('points', '-8,0 8,0 0,10')
						.attr('transform', 'translate(0, 15)');
				}

				// label group
				_timeIndicator.labelGroup = _timeIndicator.container.append('g')
					.attr('class', 'label');

				// white background
				_timeIndicator.labelGroup.append('rect')
					.attr('width', _timeIndicator.width)
					.attr('height', _fineBar.yOffset)
					.attr('rx', 3)
					.attr('ry', 3);

				// time text
				_timeIndicator.label = _timeIndicator.labelGroup.append('text')
					.attr('width', _timeIndicator.width)
					.attr('height', _fineBar.yOffset)
					.attr('x', (_timeIndicator.width / 2))
					.attr('y', 13)
					.attr('text-anchor', 'middle');

				// apply drag behavior

				_timeIndicator.container.call(d3.behavior.drag()
					.origin(Object)
					.on('dragstart', _onFineIndicatorDragStart)
					.on('drag', _onFineIndicatorDrag)
					.on('dragend', _onFineIndicatorDragEnd));
			}
		}

		function _createLiveIndicator() {
			if (_liveIndicator.created !== true) {
				_liveIndicator.created = true;

				_liveIndicator.pointerGroup = _liveIndicator.container.append('g')
					.attr('class', 'pointer');

				// tick line (drop shadows on lines apparently don't work in IE/FF, so I'm using a rect)
				_liveIndicator.pointerGroup.append('rect')
					.attr('x', -1)
					.attr('y', 0)//_fineBar.yOffset)
					.attr('width', 2)
					.attr('height', _fineBar.height);

				// label group
				_liveIndicator.labelGroup = _liveIndicator.container.append('g')
					.attr('class', 'label live')
					.attr('transform', 'translate(1)');

				// white background
				_liveIndicator.labelGroup.append('rect')
					.attr('width', _liveIndicator.width)
					.attr('height', _fineBar.yOffset);

				// time text
				_liveIndicator.label = _liveIndicator.labelGroup.append('text')
					.attr('width', _liveIndicator.width)
					.attr('height', _fineBar.yOffset)
					.attr('x', (_liveIndicator.width / 2))
					.attr('y', 12)
					.attr('text-anchor', 'middle')
					.text('LIVE');

				_liveIndicator.container.on('click', function() {
					if (_options.interactive === true) {
						_options.onTimeChange(Date.now(), 'via-live');
					}
				});
			}
		}

		function _createCoarseBar() {
			var coarseBarTop,
				coarseBarDragBehavior;

			if (_coarseBar.created !== true) {
				coarseBarTop = (_fineBar.height + _coarseBar.yOffset);

				// create configurations
				_coarseBar.configurations = [
					{
						scale: '1 day',
						formatter: _formatters.monthDay,
						ranger: d3.time.days,
						rounder: function(t) {
							return ipConfigure.core.roundDateToHours(new Date(t), 24).getTime();
						},
						milliseconds: _milliseconds.perDay,
						isKeySegment: function(d) {
							return (d.getDate() === 1);
						}
					},
					{
						scale: '12 hours',
						formatter: _formatters.minute,
						ranger: function(a, b) {
							return d3.time.hours(a, b, 12);
						},
						rounder: function(t) {
							return ipConfigure.core.roundDateToHours(new Date(t), 12).getTime();
						},
						milliseconds: _milliseconds.perHour * 12,
						isKeySegment: function(d) {
							return (d.getDate() === 1 && d.getHours() === 0);
						}
					},
					{
						scale: '6 hours',
						formatter: _formatters.minute,
						ranger: function(a, b) {
							return d3.time.hours(a, b, 6);
						},
						rounder: function(t) {
							return ipConfigure.core.roundDateToHours(new Date(t), 6).getTime();
						},
						milliseconds: _milliseconds.perHour * 6,
						isKeySegment: function(d) {
							return (d.getDate() === 1 && d.getHours() === 0);
						}
					},
					{
						scale: '3 hours',
						formatter: _formatters.minute,
						ranger: function(a, b) {
							return d3.time.hours(a, b, 3);
						},
						rounder: function(t) {
							return ipConfigure.core.roundDateToHours(new Date(t), 3).getTime();
						},
						milliseconds: _milliseconds.perHour * 3,
						isKeySegment: function(d) {
							return (d.getDate() === 1 && d.getHours() === 0);
						}
					},
					{
						scale: '1 hour',
						formatter: _formatters.minute,
						ranger: function(a, b) {
							return d3.time.hours(a, b, 1);
						},
						rounder: function(t) {
							return ipConfigure.core.roundDateToHours(new Date(t), 1).getTime();
						},
						milliseconds: _milliseconds.perHour,
						isKeySegment: function(d) {
							return (d.getHours() === 0);
						}
					},
					{
						scale: '30 minutes',
						formatter: _formatters.minute,
						ranger: function(a, b) {
							return d3.time.minutes(a, b, 30);
						},
						rounder: function(t) {
							return ipConfigure.core.roundDateToMinutes(new Date(t), 30).getTime();
						},
						milliseconds: _milliseconds.perMinute * 30,
						isKeySegment: function(d) {
							return (d.getHours() === 0 && d.getMinutes() === 0);
						}
					},
					{
						scale: '15 minutes',
						formatter: _formatters.minute,
						ranger: function(a, b) {
							return d3.time.minutes(a, b, 15);
						},
						rounder: function(t) {
							return ipConfigure.core.roundDateToMinutes(new Date(t), 15).getTime();
						},
						milliseconds: _milliseconds.perMinute * 15,
						isKeySegment: function(d) {
							return (d.getHours() === 0 && d.getMinutes() === 0);
						}
					},
					{
						scale: '5 minutes',
						formatter: _formatters.minute,
						ranger: function(a, b) {
							return d3.time.minutes(a, b, 5);
						},
						rounder: function(t) {
							return ipConfigure.core.roundDateToMinutes(new Date(t), 5).getTime();
						},
						milliseconds: _milliseconds.perMinute * 5,
						isKeySegment: function(d) {
							return (d.getHours() === 0 && d.getMinutes() === 0);
						}
					},
					{
						scale: '1 minute',
						formatter: _formatters.minute,
						ranger: function(a, b) {
							return d3.time.minutes(a, b, 1);
						},
						rounder: function(t) {
							return ipConfigure.core.roundDateToMinutes(new Date(t), 1).getTime();
						},
						milliseconds: _milliseconds.perMinute,
						isKeySegment: function(d) {
							return (d.getHours() === 0 && d.getMinutes() === 0);
						}
					}
				];

				// create an offset container for coarse bar segments and context
				_coarseBar.coarseContainer = _coarseBar.container.append('g')
					.attr('class', 'coarse-container')
					.attr('transform', 'translate(0, ' + coarseBarTop + ')');

				// create a background rect for the container
				_coarseBar.coarseContainer.append('rect')
					.attr('class', 'coarse-background')
					.attr('width', '100%')
					.attr('height', (_coarseBar.height - _coarseBar.yOffset) + 'px');

				// create a group to contain the time blocks
				_coarseBar.segmentContainer = _coarseBar.coarseContainer.append('g')
					.attr('class', 'segment-container')
					.data([
						{ x: -_coarseBar.segmentWidth }
					])
					.attr('transform', function(d) {
						return 'translate(' + d.x + ')';
					});

				// create the live indicator
				_coarseBar.liveBar.container = _coarseBar.container.append('g')
					.attr('class', 'live inactive')
					.on('click', _onCoarseLiveBarClick);

				_coarseBar.liveBar.rect = _coarseBar.liveBar.container
					.append('rect');

				// create the time window
				_coarseBar.timeWindowMask = _coarseBar.container.append('g')
					.attr('class', 'time-window-mask');

				_coarseBar.timeWindow = _coarseBar.container.append('g')
					.attr('class', 'time-window');

				_coarseBar.timeWindowCenter = _coarseBar.timeWindow.append('rect')
					.attr('class', 'time-window-center')
					.attr('x', 0)
					.attr('y', coarseBarTop)
					.attr('height', _coarseBar.height);

				_coarseBar.timeWindowLeftHandle = _coarseBar.timeWindow.append('rect')
					.attr('class', 'time-window-left-handle')
					.attr('x', -15)
					.attr('y', coarseBarTop)
					.attr('width', _coarseBar.timeWindowHandleWidth)
					.attr('height', _coarseBar.height);

				_coarseBar.timeWindowRightHandle = _coarseBar.timeWindow.append('rect')
					.attr('class', 'time-window-right-handle')
					.attr('y', coarseBarTop)
					.attr('width', _coarseBar.timeWindowHandleWidth)
					.attr('height', _coarseBar.height);

				coarseBarDragBehavior = d3.behavior.drag()
					.origin(function() {
						return _coarseBar.segmentContainer.data()[0];
					})
					.on('dragstart', _onCoarseBarDragStart)
					.on('drag', _onCoarseBarDrag)
					.on('dragend', _onCoarseBarDragEnd);

				_coarseBar.leftMask = _coarseBar.timeWindowMask.append('polygon')
					.attr('class', 'time-window-left-mask')
					.attr('points', '0,' + coarseBarTop + ' 9000,' + coarseBarTop + ' 9015,' + (coarseBarTop + 15) + '  9015,' + (coarseBarTop + _coarseBar.height) + ' 0,' + (coarseBarTop + 25))
					.on('click', _onCoarseBarClick)
					.call(coarseBarDragBehavior);

				_coarseBar.rightMask = _coarseBar.timeWindowMask.append('polygon')
					.attr('class', 'time-window-right-mask')
					.attr('points', '15,' + coarseBarTop + ' 9000,' + coarseBarTop.toString() + ' 9000,' + (coarseBarTop + _coarseBar.height) + ' 0,' + (coarseBarTop + _coarseBar.height) + ' 0,' + (coarseBarTop + 15))
					.on('click', _onCoarseBarClick)
					.call(coarseBarDragBehavior);

				// this sucks, but seems the most efficient way to create a crude drop shadow around the edge of the time window
				_coarseBar.leftMaskShadow = _coarseBar.timeWindowMask.append('g')
					.attr('class', 'time-window-mask-left-shadow');

				_coarseBar.leftMaskShadow.append('line').attr('y1', 63).attr('x1', 0).attr('y2', 63).attr('x2', 9000).attr('stroke-width', 1).attr('stroke', 'black').attr('shape-rendering', 'crispEdges').attr('opacity', 0.33);
				_coarseBar.leftMaskShadow.append('line').attr('y1', 64).attr('x1', 0).attr('y2', 64).attr('x2', 9001).attr('stroke-width', 1).attr('stroke', 'black').attr('shape-rendering', 'crispEdges').attr('opacity', 0.25);
				_coarseBar.leftMaskShadow.append('line').attr('y1', 65).attr('x1', 0).attr('y2', 65).attr('x2', 9002).attr('stroke-width', 1).attr('stroke', 'black').attr('shape-rendering', 'crispEdges').attr('opacity', 0.10);
				_coarseBar.leftMaskShadow.append('line').attr('y1', 64).attr('x1', 9000).attr('y2', 78).attr('x2', 9014).attr('stroke-width', 1).attr('stroke', 'black').attr('shape-rendering', 'crispEdges').attr('opacity', 0.33);
				_coarseBar.leftMaskShadow.append('line').attr('y1', 64).attr('x1', 8999).attr('y2', 79).attr('x2', 9014).attr('stroke-width', 1).attr('stroke', 'black').attr('shape-rendering', 'crispEdges').attr('opacity', 0.25);
				_coarseBar.leftMaskShadow.append('line').attr('y1', 79).attr('x1', 9014).attr('y2', 88).attr('x2', 9014).attr('stroke-width', 1).attr('stroke', 'black').attr('shape-rendering', 'crispEdges').attr('opacity', 0.33);

				_coarseBar.rightMaskShadow = _coarseBar.timeWindowMask.append('g')
					.attr('class', 'time-window-mask-right-shadow');

				_coarseBar.rightMaskShadow.append('line').attr('y1', 63).attr('x1', 15).attr('y2', 63).attr('x2', 9000).attr('stroke-width', 1).attr('stroke', 'black').attr('shape-rendering', 'crispEdges').attr('opacity', 0.33);
				_coarseBar.rightMaskShadow.append('line').attr('y1', 64).attr('x1', 14).attr('y2', 64).attr('x2', 9001).attr('stroke-width', 1).attr('stroke', 'black').attr('shape-rendering', 'crispEdges').attr('opacity', 0.25);
				_coarseBar.rightMaskShadow.append('line').attr('y1', 65).attr('x1', 13).attr('y2', 65).attr('x2', 9002).attr('stroke-width', 1).attr('stroke', 'black').attr('shape-rendering', 'crispEdges').attr('opacity', 0.10);
				_coarseBar.rightMaskShadow.append('line').attr('y1', 65).attr('x1', 14).attr('y2', 78).attr('x2', 1).attr('stroke-width', 1).attr('stroke', 'black').attr('shape-rendering', 'crispEdges').attr('opacity', 0.33);
				_coarseBar.rightMaskShadow.append('line').attr('y1', 65).attr('x1', 15).attr('y2', 79).attr('x2', 1).attr('stroke-width', 1).attr('stroke', 'black').attr('shape-rendering', 'crispEdges').attr('opacity', 0.25);
				_coarseBar.rightMaskShadow.append('line').attr('y1', 79).attr('x1', 1).attr('y2', 88).attr('x2', 1).attr('stroke-width', 1).attr('stroke', 'black').attr('shape-rendering', 'crispEdges').attr('opacity', 0.33);

				// wire up time window drag events
				_coarseBar.timeWindow.data([
						{ x: 0, width: 0 }
					])
					.call(d3.behavior.drag()
						.on('dragstart', _onTimeWindowDragStart)
						.on('drag', _onTimeWindowDrag)
						.on('dragend', _onTimeWindowDragEnd));

				// create a fine timeline resize handle on the right edge
				_fineBar.rightHandleGrabber = _fineBar.sizerGroup.append('g')
					.style('cursor', 'e-resize')
					.attr('transform', 'translate(' + (_width - 25) + ')')
					.data([
						{ x: _width - 25 }
					])
					.call(d3.behavior.drag()
						.origin(Object)
						.on('dragstart', _onFineRightHandleDragStart)
						.on('drag', _onFineRightHandleDrag)
						.on('dragend', _onFineRightHandleDragEnd));

				_fineBar.rightHandle = _fineBar.rightHandleGrabber.append('rect')
					.attr('x', 0)
					.attr('y', _fineBar.yOffset)
					.attr('width', 25)
					.attr('height', _fineBar.height - _fineBar.yOffset)
					.attr('fill', '#0f313d')
					.attr('opacity', '0.5');

				_fineBar.rightHandleGrabber.append('line').attr('x1', 9.5).attr('x2', 9.5).attr('y1', _fineBar.yOffset + 15).attr('y2', _fineBar.yOffset + 15 + 18).attr('stroke', '#c8e3f1');
				_fineBar.rightHandleGrabber.append('line').attr('x1', 12.5).attr('x2', 12.5).attr('y1', _fineBar.yOffset + 15).attr('y2', _fineBar.yOffset + 15 + 18).attr('stroke', '#c8e3f1');
				_fineBar.rightHandleGrabber.append('line').attr('x1', 15.5).attr('x2', 15.5).attr('y1', _fineBar.yOffset + 15).attr('y2', _fineBar.yOffset + 15 + 18).attr('stroke', '#c8e3f1');

				// create a fine timeline resize handle on the left edge
				_fineBar.leftHandleGrabber = _fineBar.sizerGroup.append('g')
					.style('cursor', 'e-resize')
					.call(d3.behavior.drag()
						.origin(Object)
						.on('dragstart', _onFineLeftHandleDragStart)
						.on('drag', _onFineLeftHandleDrag)
						.on('dragend', _onFineLeftHandleDragEnd))
					.data([
						{ x: 25 }
					]);

				_fineBar.leftHandle = _fineBar.leftHandleGrabber.append('rect')
					.attr('x', 0)
					.attr('y', _fineBar.yOffset)
					.attr('width', 25)
					.attr('height', _fineBar.height - _fineBar.yOffset)
					.attr('fill', '#0f313d')
					.attr('opacity', '0.5');

				_fineBar.leftHandlGrabberLineGroup = _fineBar.leftHandleGrabber.append('g')
					.attr('transform', 'translate(9)');

				_fineBar.leftHandlGrabberLineGroup.append('line').attr('x1', 0.5).attr('x2', 0.5).attr('y1', _fineBar.yOffset + 15).attr('y2', _fineBar.yOffset + 15 + 18).attr('stroke', 'white');
				_fineBar.leftHandlGrabberLineGroup.append('line').attr('x1', 3.5).attr('x2', 3.5).attr('y1', _fineBar.yOffset + 15).attr('y2', _fineBar.yOffset + 15 + 18).attr('stroke', 'white');
				_fineBar.leftHandlGrabberLineGroup.append('line').attr('x1', 6.5).attr('x2', 6.5).attr('y1', _fineBar.yOffset + 15).attr('y2', _fineBar.yOffset + 15 + 18).attr('stroke', 'white');

				// adjust the height so that the coarse bar will be visible
				_container.style('height', (_fineBar.height + _coarseBar.height) + 'px');

				// don't let this ever happen again
				_coarseBar.created = true;

				// initialize these so we can write cleaner code
				_coarseBar.dragging = false;
				_coarseBar.timeWindow.dragging = false;
			}
		}

		// position functions

		function _positionFineIndicator(x) {
			var halfWidth = _timeIndicator.width / 2,
				labelOffset = -halfWidth,
				calculatedCurrentWidth = _calculateCurrentWidth(),
				display = '',
				fineBarLeftOffset;

			// position the indicator
			_timeIndicator.container.data([{ x: x }])
				.attr('transform', function(d) {
					return 'translate(' + Math.round(d.x + _fineBar.xOffset) + ')';
				});

			// prevent the label from going off the page
			if ((x + _fineBar.xOffset) < halfWidth) {
				labelOffset = (halfWidth - (x + _fineBar.xOffset) - halfWidth);

				if (labelOffset > -_fineBar.xOffset) {
					display = 'none';
				}
			}
			else {
				if ((x + _fineBar.xOffset) > (_width - halfWidth)) {
					labelOffset = (-halfWidth + -(halfWidth - (_width - (x + _fineBar.xOffset))));
				}

				if (x > calculatedCurrentWidth) {
					display = 'none';
				}
			}

			// prevent the label from overlapping the live indicator
			if (x + _timeIndicator.width / 2 - _fineBar.xOffset > calculatedCurrentWidth - (_liveIndicator.width)) {
				// handle when the live indicator is anchored on the right edge
				labelOffset = -((x + _timeIndicator.width / 2 + _liveIndicator.width) - (calculatedCurrentWidth)) - _liveIndicator.width + _fineBar.xOffset - 9;
			}
			else {
				// take into account left/right resize bars that appear in playback mode
				fineBarLeftOffset = _mode === 'live' ? 0 : 25;

				// handle when the live indicator is close to the time indicator
				// querying here because sometimes things get updated out of order, so sometimes (_fineBar.scale(Date.now())) isn't where the label is
				if ((_liveIndicator.labelGroup.select('rect').node(0).getBoundingClientRect().left - _xOffset - 1) - fineBarLeftOffset < x + _timeIndicator.width / 2) {
					labelOffset = -((x + _timeIndicator.width / 2) - _fineBar.scale(Date.now())) - (_timeIndicator.width / 2);
				}
			}

			// set or remove the live class on the labelGroup (to get better live indication)
			if (isNaN(_currentTime) === false) {
				_liveIndicator.labelGroup.attr('class', ((Date.now() - _currentTime < 5000) ? 'label live' : 'label'));
			}

			_timeIndicator.pointerGroup.style('display', display);
			_timeIndicator.labelGroup.attr('transform', 'translate(' + Math.round(labelOffset) + ')');
		}

		function _positionFineLiveIndicator() {
			var x,
				calculatedCurrentWidth,
				liveBar = {
					containerPosition: 0,
					rectWidth: _width
				},
				liveIndicator = {
					containerPosition: _fineBar.xOffset,
					labelGroupPosition: 0,
					display: 'none'
				};

			x = _fineBar.scale(Date.now());

			if (x >= 0) {
				// if "now" is visible on the timeline, or to the right of it, override the defaults
				calculatedCurrentWidth = _calculateCurrentWidth();

				// position and size the live bar
				liveBar.containerPosition = (x + _fineBar.xOffset);
				liveBar.rectWidth = calculatedCurrentWidth;

				// position the live indicator
				liveIndicator.display = '';
				liveIndicator.containerPosition = Math.round(x + _fineBar.xOffset);

				// if the live indicator is near the edge, calculate labelGroup offset
				if (x > calculatedCurrentWidth - _liveIndicator.width + _fineBar.xOffset) {
					liveIndicator.labelGroupPosition = -Math.round(_liveIndicator.width - (calculatedCurrentWidth - x) - _fineBar.xOffset);
				}

				// hide the live indicator pointer if now is off the right edge
				if (x > calculatedCurrentWidth) {
					liveIndicator.display = 'none';
				}
			}

			// position the live bar
			_fineBar.liveBar.container
				.attr('transform', 'translate(' + liveBar.containerPosition + ', ' + _fineBar.yOffset + ')');

			_fineBar.liveBar.rect.attr('width', liveBar.rectWidth);

			_liveIndicator.pointerGroup
				.style('display', liveIndicator.display);

			// this conditional exists because of a bug in firefox (https://bugzilla.mozilla.org/show_bug.cgi?id=1217012)
			// the else condition is all that's really required.. when the bug is fix we can fall back to those 2 statements
			if (liveIndicator.display === 'none') {
				// if the flag pole is hidden, the flag should be anchored to the right of the screen
				_liveIndicator.container
					.attr('transform', 'translate(' + (_width - _liveIndicator.width) + ')');

				_liveIndicator.labelGroup
					.attr('transform', 'translate(0)');
			}
			else {
				// if the flag pole is visible, position the flag as calculated above
				_liveIndicator.container
					.attr('transform', 'translate(' + liveIndicator.containerPosition + ')');

				_liveIndicator.labelGroup
					.attr('transform', 'translate(' + ((liveIndicator.labelGroupPosition === 0) ? 1 : liveIndicator.labelGroupPosition) + ')');
			}
		}

		function _positionCoarseLiveIndicator() {
			if (_coarseBar.created === true && _mode === 'playback') {
				var x = _coarseBar.scale(Date.now());

				if (x < 0) {
					// if "now" is off the left edge of the screen, just fill the viewable area with the live bar
					_coarseBar.liveBar.rect
						.attr('width', _width)
						.attr('height', _fineBar.height - _fineBar.yOffset);

					_coarseBar.liveBar.container
						.attr('transform', 'translate(0, ' + (_fineBar.height + _coarseBar.yOffset) + ')')
						.attr();
				}
				else {
					if (x < _width) {
						_coarseBar.liveBar.rect
							.attr('width', _width)
							.attr('height', _fineBar.height - _fineBar.yOffset);

						_coarseBar.liveBar.container
							.data([
								{
									x: x
								}
							])
							.attr('class', 'live')
							.attr('transform', 'translate(' + x + ', ' + (_fineBar.height + _coarseBar.yOffset) + ')')
							.attr();
					}
					else {
						_coarseBar.liveBar.container
							.data([
								{
									x: (_width - 25)
								}
							])
							.attr('class', 'live inactive')
							.attr('transform', 'translate(' + (_width - 25) + ', ' + (_fineBar.height + _coarseBar.yOffset) + ')');
					}
				}
			}
		}

		function _updateFineIndicatorLabelFromPoint(x) {
			// update the indicator's label
			_timeIndicator.label.text(_timeIndicator.indicatorLabelFormat(_fineBar.scale.invert(x)));
		}

		function _updateFineIndicatorLabelFromTime(time) {
			// update the indicator's label
			_timeIndicator.label.text(_timeIndicator.indicatorLabelFormat(new Date(time)));
		}

		function _positionTimeWindow(d, animate) {
			var timeWindowSubject = animate ? _coarseBar.timeWindow.transition() : _coarseBar.timeWindow,
				timeWindowCenterSubject = animate ? _coarseBar.timeWindowCenter.transition() : _coarseBar.timeWindowCenter,
				timeWindowRightHandleSubject = animate ? _coarseBar.timeWindowRightHandle.transition() : _coarseBar.timeWindowRightHandle,
				leftMaskSubject = animate ? _coarseBar.leftMask.transition() : _coarseBar.leftMask,
				leftMaskShadowSubject = animate ? _coarseBar.leftMaskShadow.transition() : _coarseBar.leftMaskShadow,
				rightMaskShadowSubject = animate ? _coarseBar.rightMaskShadow.transition() : _coarseBar.rightMaskShadow,
				rightMaskSubject = animate ? _coarseBar.rightMask.transition() : _coarseBar.rightMask;

			timeWindowSubject.attr('transform', 'translate(' + d.x + ')');
			timeWindowCenterSubject.attr('width', d.width);
			timeWindowRightHandleSubject.attr('x', d.width - 5);

			leftMaskSubject.attr('transform', 'translate(' + (d.x - 9000 - 16) + ')');
			leftMaskShadowSubject.attr('transform', 'translate(' + (d.x - 9000 - 16) + ')');
			rightMaskSubject.attr('transform', 'translate(' + (d.x + d.width) + ')');
			rightMaskShadowSubject.attr('transform', 'translate(' + (d.x + d.width) + ')');
		}

		function _positionFineIndicatorAfterAdjustment(via) {
			var indicatedTime = _currentTime;

			if (indicatedTime < _fineBar.start || indicatedTime > _fineBar.stop) {
				// the user has dragged the current time out of view, yank it back to the left
				_options.onTimeChange(_fineBar.start, via);
			}
		}

		function _d3TimeOrdinalSuffix(number) {
			var suffix = 'th',
				tail = number % 100;

			if (tail < 11 || tail > 13) {
				switch (tail % 10) {
					case 1:
						suffix = 'st';
						break;
					case 2:
						suffix = 'nd';
						break;
					case 3:
						suffix = 'rd';
						break;
					default:
						break;
				}
			}

			return suffix;
		}

		// render functions

		function _fineRenderTimeline(animate, completeCallback) {
			if (_fineBar.animating !== true) {
				// adjust the scale w/ the new domain
				if (_fineBar.created === true) {
					_fineBar.scale.domain([new Date(_fineBar.start), new Date(_fineBar.stop)]);

					_formatters.fineChecksum = {};

					if (animate) {
						_fineBar.animating = true;

						// clear the fake minor ticks
						_fineClearEnhancements();

						// redraw the axis
						_fineBar.axisGroup
							.transition()
							.duration(250)
							.call(_fineBar.axis)
							.each('end', function() {
								setTimeout(_fineRenderEnhancements);

								_fineBar.animating = false;

								if (completeCallback) {
									completeCallback();
								}
							});
					}
					else {
						// redraw the axis
						_fineBar.axisGroup
							.call(_fineBar.axis);

						// redraw the fake minor ticks
						_fineRenderEnhancements();

						if (completeCallback) {
							completeCallback();
						}
					}
				}
			}
		}

		function _fineClearEnhancements() {
			_fineBar.axisMinorGroup.selectAll('line').remove();
			_fineBar.axisMotionGroup.selectAll('rect').remove();
		}

		function _debounceFineRenderEnhancements() {
			_fineBar.debounceRenderEnhancementTimer = setTimeout(function() {
				if (_fineBar.animating !== true) {
					_fineClearEnhancements();
					_fineRenderEnhancements();
				}
			}, 50);
		}

		function _fineRenderEnhancements() {
			var minorTickStep = 6,
				tickData = [],
				minorTickPositions = [],
				i,
				j,
				minorTicks,
				motionRects,
				now,
				boldMonthChecksum = {};

			if (_fineBar.debounceRenderEnhancementTimer) {
				clearTimeout(_fineBar.debounceRenderEnhancementTimer);
				delete _fineBar.debounceRenderEnhancementTimer;
			}

			// collect tick information generated by d3.axis so that we can expand on it
			_fineBar.axisGroup.selectAll('g').data().forEach(function(d) {
				tickData.push({
					date: d,
					position: _fineBar.scale(d)
				});
			});

			if (tickData.length > 0) {
				// since there's no guarantee that the data will be sorted, we need to sort it now
				tickData.sort(function(a, b) {
					return a.position - b.position;
				});

				// create ticks from the beginning to the first tick (in reverse order for coding ease)
				for (j = tickData[0].position - minorTickStep; j > 0; j -= minorTickStep) {
					minorTickPositions.push(j);
				}

				// create all the inner ticks
				for (i = 0; i < tickData.length - 1; i += 1) {
					for (j = tickData[i].position + minorTickStep; j < tickData[i + 1].position - (minorTickStep / 2); j += minorTickStep) {
						minorTickPositions.push(j);
					}
				}

				// create ticks from the last tick to the end
				for (j = tickData[tickData.length - 1].position + minorTickStep; j < _calculateCurrentWidth(); j += minorTickStep) {
					minorTickPositions.push(j);
				}

				// create/update minor ticks
				minorTicks = _fineBar.axisMinorGroup.selectAll('line').data(minorTickPositions);

				minorTicks.exit().remove();
				minorTicks.enter().append('line');

				_fineBar.axisMinorGroup.selectAll('line')
					.attr('y2', 6)
					.attr('x2', 0)
					.attr('transform', function(d) {
						return 'translate(' + d + ')';
					});

				// bold midnight
				_fineBar.axisGroup.selectAll('text')
					.style('font-weight', function(d) {
						var key = d.getMonth().toString() + d.getDate().toString() + d.getHours().toString() + d.getMinutes().toString(),
							weight = '';

						if (_formatters.fineChecksum[key] === true) {
							if (boldMonthChecksum[key] !== true) {
								boldMonthChecksum[key] = true;

								weight = 'bold';
							}
						}

						return weight;
					});

				// create/update motion indication bars
				motionRects = _fineBar.axisMotionGroup.selectAll('rect').data(_motionEventData);
				motionRects.exit().remove();
				motionRects.enter().append('rect')
					.attr('height', 7);

				now = Date.now();

				_fineBar.axisMotionGroup.selectAll('rect')
					.attr('y', (_motionEventsAnchored) ? 0 : 35)
					.attr('x', function(slice) {
						return _fineBar.scale(slice.start + window.clientServerTime.diff);
					})
					.attr('width', function(slice) {
						if (slice.start + window.clientServerTime.diff > now) {
							// set to zero if start is in the future
							return 0;
						}
						else if (slice.stop + window.clientServerTime.diff > now) {
							// set to now if the stop is in the future
							return _fineBar.scale(now) - _fineBar.scale(slice.start + window.clientServerTime.diff);
						}
						else {
							return _fineBar.scale(slice.stop + window.clientServerTime.diff) - _fineBar.scale(slice.start + window.clientServerTime.diff) + 1;
						}
					})
					.attr('fill', '#f55');

				// update text elements to contain time with visual differences
				//_fineBar.axisGroup.selectAll('text').attr('fill', function(d, i) {
				//	return i % 2 === 0 ? '#000' : '#888';
				//});
			}
		}

		function _coarseRenderTimeline() {
			var currentConfiguration = _coarseBar.configurations[_coarseBar.configurationIndex],
				segmentCount = Math.ceil(_width / _coarseBar.segmentWidth) + 2,
				range,
				segments,
				segmentsEnter;

			// calculate the time at which the coarse bar ends
			_coarseBar.stopTime = _coarseBar.startTime + segmentCount * currentConfiguration.milliseconds;

			// update the scale function
			_coarseBar.scale.domain([new Date(_coarseBar.startTime), new Date(_coarseBar.stopTime)]);

			// raise an event so that user knows that the coarse bar has been adjusted
			_options.onCoarseTimeChange(_coarseBar.startTime, _coarseBar.stopTime);

			// generate new data for the segments
			range = currentConfiguration.ranger(new Date(_coarseBar.startTime), new Date(_coarseBar.stopTime));

			// update data
			segments = _coarseBar.segmentContainer
				.selectAll('g')
				.data(range);

			// apply data to child elements and update necessary properties
			segments.select('rect');

			segments.select('text')
				.text(currentConfiguration.formatter);

			// remove old janks (should only happen on resize)
			segments.exit().remove();

			// create new janks (should happen initially, or on resize/drag)
			segmentsEnter = segments.enter()
				.append('g');

			segmentsEnter
				.append('rect')
				.attr('width', _coarseBar.segmentWidth)
				.attr('height', _coarseBar.height);

			segmentsEnter
				.append('text')
				.attr('x', _coarseBar.segmentWidth / 2)
				.attr('y', 18)
				.attr('text-anchor', 'middle')
				.text(currentConfiguration.formatter);

			// update segments
			segments
				.attr('transform', function(d, i) {
					return 'translate(' + i * (_coarseBar.segmentWidth) + ')';
				});
		}

		// etc

		function _calculateCurrentWidth() {
			return ((_mode === 'live') ? _width : _width - 50);
		}

		function _adjustScaleComplete(atSegment, newConfigurationIndex, beSoSticky, savedTimeWindowPosition) {
			var timeWindowData = _coarseBar.timeWindow.data()[0],
				temp = {},
				i;

			_coarseBarAdjustResolution(atSegment, newConfigurationIndex);
			_coarseRenderTimeline();
			_updateCoarseScaleRange(_coarseBar.segmentContainer.data()[0].x);

			if (beSoSticky === true) {
				// the same start position may not be accessible (by pixel) with the new coarse bar scale, so
				// we should adjust the start position wider until that previous start is encompassed
				i = 0;
				temp.x = Math.round(_coarseBar.scale(savedTimeWindowPosition.start));
				temp.start = _coarseBar.scale.invert(temp.x).getTime();
				while (temp.start > savedTimeWindowPosition.start) {
					temp.start = _coarseBar.scale.invert(temp.x - i).getTime();
					i += 1;
				}
				savedTimeWindowPosition.start = temp.start;

				// maintain that minimum 5px width
				if (_coarseBar.scale(savedTimeWindowPosition.stop) - _coarseBar.scale(savedTimeWindowPosition.start) < _coarseBar.timeWindowMinimumWidth) {
					savedTimeWindowPosition.stop = _coarseBar.scale.invert(_coarseBar.scale(savedTimeWindowPosition.start) + _coarseBar.timeWindowMinimumWidth).getTime();
				}

				_setTimeWindowPosition(savedTimeWindowPosition.start, savedTimeWindowPosition.stop, true);
			}

			_fineBar.start = _coarseBar.scale.invert(timeWindowData.x).getTime();
			_fineBar.stop = _coarseBar.scale.invert(timeWindowData.x + timeWindowData.width).getTime();
			_fineRenderTimeline(true);

			options.onScaleAdjust(newConfigurationIndex);
		}

		function _calculateTicksForWidth() {
			var ticks;

			if (_width < 300) {
				ticks = 2;
			}
			else if (_width < 900) {
				ticks = 5;
			}
			else {
				ticks = 10;
			}

			return ticks;
		}

		function _updateCoarseScaleRange(currentX) {
			var segmentCount = Math.ceil(_width / _coarseBar.segmentWidth) + 2,
				scaleLower = currentX + Math.round(_coarseBar.segmentWidth / 2),
				scaleUpper = (_coarseBar.segmentWidth * segmentCount) + scaleLower;

			_coarseBar.scale
				.range([scaleLower, scaleUpper]);
		}

		function _coarseBarAdjustResolution(atSegment, newConfigurationIndex) {
			var currentConfiguration = _coarseBar.configurations[_coarseBar.configurationIndex],
				newConfiguration = _coarseBar.configurations[newConfigurationIndex],
				adjustment,
				normalizedStartTime,
				multiplier;

			// TODO rewrite this entirely so that the timeWindow remains wrapped around where the needle is so now playback adjustments occur

			// resolution has been adjusted, make the change to the configurationIndex now
			_coarseBar.configurationIndex = newConfigurationIndex;

			if (currentConfiguration.milliseconds > newConfiguration.milliseconds) {
				// calculate the multiplier to supprt skipping over multipliers (jumping from config[0] to config[2]
				multiplier = currentConfiguration.milliseconds / newConfiguration.milliseconds - 1;

				// if we're increasing the resolution of the coarse bar, we need to multiply
				// the segment we're at by the number of milliseconds per segment
				adjustment = atSegment * newConfiguration.milliseconds * multiplier;

				// and make the appropriate adjustment to both sides
				_coarseBar.startTime += adjustment;
			}
			else {
				// calculate the multiplier to supprt skipping over multipliers (jumping from config[2] to config[0]
				multiplier = newConfiguration.milliseconds / currentConfiguration.milliseconds - 1;

				// if we're decreasing resolution, we need to multiply by the previous config's value
				adjustment = atSegment * currentConfiguration.milliseconds * multiplier;

				// and make the appropriate adjustment to both sides
				_coarseBar.startTime -= adjustment;
			}

			// normalize startTime and stopTime to be at the right scale. this probably isn't the best approach, but
			// it seems to work for now.  the time window shifts over time when the user zooms in and out over and over.
			normalizedStartTime = newConfiguration.rounder(_coarseBar.startTime);

			if (normalizedStartTime - _coarseBar.startTime !== 0) {
				//stopTime += (normalizedStartTime - startTime);
				_coarseBar.startTime = normalizedStartTime;
			}
		}

		function _coarseBarSplit(atSegment, completeCallback) {
			var currentValue,
				newValue,
				leftCounter = 0,
				rightCounter = 0;

			_coarseBar.segmentContainer.selectAll('g').each(function(data, index) {
				var element = d3.select(this);

				currentValue = (index * _coarseBar.segmentWidth);

				if (atSegment > index) {
					newValue = currentValue - (atSegment - index) * _coarseBar.segmentWidth;

					// send these guys left
					element
						.transition()
						.duration(200)
						.each(function() {
							++leftCounter;
						})
						.each('end', function() {
							--leftCounter;

							if (rightCounter + leftCounter === 0) {
								completeCallback();
							}
						})
						.attr('transform', 'translate(' + newValue + ')');
				}
				else if (atSegment < index) {
					newValue = currentValue + (index - atSegment) * _coarseBar.segmentWidth;

					// send these guys right
					element
						.transition()
						.duration(200)
						.each(function() {
							++rightCounter;
						})
						.each('end', function() {
							--rightCounter;

							if (rightCounter + leftCounter === 0) {
								completeCallback();
							}
						})
						.attr('transform', 'translate(' + newValue + ')');
				}
			});
		}

		function _coarseBarUnsplit(atSegment, completeCallback) {
			var elements = _coarseBar.segmentContainer.selectAll('g'),
				index,
				offset,
				currentValue,
				newValue,
				leftCounter = 0,
				rightCounter = 0,
				eachIteration,
				eachEnd;

			// move right bits left
			eachIteration = function() {
				++rightCounter;
			};

			eachEnd = function() {
				--rightCounter;

				if (rightCounter + leftCounter === 0) {
					completeCallback();
				}
			};

			offset = 1;

			for (index = atSegment + 2; index < elements[0].length; index += 2) {
				currentValue = (index * _coarseBar.segmentWidth);
				newValue = currentValue - (offset * _coarseBar.segmentWidth);

				d3.select(elements[0][index])
					.transition()
					.duration(200)
					.each(eachIteration)
					.each('end', eachEnd)
					.attr('transform', 'translate(' + newValue + ')');

				offset++;
			}

			// move left bits right
			eachIteration = function() {
				++leftCounter;
			};

			eachEnd = function() {
				--leftCounter;

				if (rightCounter + leftCounter === 0) {
					completeCallback();
				}
			};

			offset = 1;

			// because svg calculates z-index based on how elements were rendered, we have to hide some bits otherwise stuff gets overlaid
			for (index = atSegment - 1; index >= 0; index -= 2) {
				d3.select(elements[0][index]).attr('transform', 'translate(-' + _coarseBar.segmentWidth + ')');
			}

			for (index = atSegment - 2; index >= 0; index -= 2) {
				currentValue = (index * _coarseBar.segmentWidth);
				newValue = currentValue + (offset * _coarseBar.segmentWidth);

				d3.select(elements[0][index])
					.transition()
					.duration(200)
					.each(eachIteration)
					.each('end', eachEnd)
					.attr('transform', 'translate(' + newValue + ')');

				offset++;
			}
		}

		function _setTimeWindowPosition(startDate, stopDate, animate) {
			var d,
				allowableWidth;

			if (_coarseBar.created === true) {
				if (stopDate - startDate === 0) {
					stopDate = _coarseBar.scale.invert(_coarseBar.scale(startDate) - 5);
				}

				d = _coarseBar.timeWindow.data()[0];

				if (startDate <= stopDate) {
					allowableWidth = (_width - _coarseBar.timeWindowTriangleWidth);

					// update the left edge of the time window based on the scale function
					d.x = Math.round(_coarseBar.scale(startDate));

					// constrain time window within bounds of coarse bar
					d.x = Math.max(_coarseBar.timeWindowTriangleWidth, d.x);
					d.x = Math.min(allowableWidth, d.x);

					// update the width
					d.width = Math.round(_coarseBar.scale(stopDate) - d.x);

					// and constrain it
					d.width = Math.max(0, d.width);
					d.width = Math.min(allowableWidth - d.x, d.width);

					// and position it
					_positionTimeWindow(d, animate);
				}
			}
		}

		function _setupCoarseBarAtTime() {
			var segmentCount,
				i,
				fineBarLength,
				fineBarSpansThisManySegments,
				multiplier;

			if (_coarseBar.created === true) {
				// calculate which configuration index we should use based on _fineBar.stop - _fineBar.start
				segmentCount = Math.ceil(_width / _coarseBar.segmentWidth) + 2;
				fineBarLength = _fineBar.stop - _fineBar.start;
				i = _coarseBar.configurations.length;

				while (i--) {
					// determine how many segments fineBarLength will scale
					fineBarSpansThisManySegments = fineBarLength / _coarseBar.configurations[i].milliseconds;

					if (fineBarSpansThisManySegments <= 0.25) {
						_coarseBar.configurationIndex = i;
						break;
					}
				}

				// let subscribers know the scale was adjusted
				options.onScaleAdjust(_coarseBar.configurationIndex);

				// attempt to center where the time window will land over the coarse bar
				multiplier = (segmentCount - 2) / 2;
				_coarseBar.startTime = _fineBar.start - multiplier * _coarseBar.configurations[_coarseBar.configurationIndex].milliseconds + (fineBarSpansThisManySegments * _coarseBar.configurations[_coarseBar.configurationIndex].milliseconds / 2);

				// normalize startTime
				_coarseBar.startTime = ipConfigure.core.roundDateToMinutes(new Date(_coarseBar.startTime), 1).getTime();

				// force the coarse bar to render
				_coarseRenderTimeline();
				_updateCoarseScaleRange(_coarseBar.segmentContainer.data()[0].x);

				// position the time window
				_setTimeWindowPosition(_fineBar.start, _fineBar.stop);
			}
		}

		// initialization

		function _init() {
			_log('timeline options', _options);

			_createSvgContainer();
		}

		// public properties

		that.options = _options;

		// public methods

		that.getTime = function() {
			return _currentTime;
		};

		that.setTime = function(time) {
			if (_fineBar.animating !== true) {
				_currentTime = time;

				if (_mode === 'live' && _fineBar.dragging !== true && _timeIndicator.dragging !== true) {
					// if we're tracking live and the user isn't dragging the fine bar, update the fine timeline
					_fineBar.start = time - _fineBar.liveHalfTime;
					_fineBar.stop = time + _fineBar.liveHalfTime;

					_fineRenderTimeline();
				}

				_positionFineLiveIndicator();

				if (_mode === 'playback') {
					_positionCoarseLiveIndicator();
				}

				if ((_timeIndicator.dragging === false && _fineBar.dragging === false) || (_mode === 'playback' && _fineBar.dragging === true)) {
					_positionFineIndicator(Math.round(_fineBar.scale(new Date(time))));
					_updateFineIndicatorLabelFromTime(time);
				}

				// raise onEndReached
				if (_coarseBar.created === true && _fineBar.dragging === false && _coarseBar.dragging === false && _coarseBar.timeWindow.dragging === false && (time >= _fineBar.stop || time < _fineBar.start)) {
					_options.onEndReached([_fineBar.start, _fineBar.stop]);
				}
			}
		};

		that.getCoarseTime = function() {
			return [_coarseBar.startTime, _coarseBar.stopTime];
		};

		that.setCoarseTime = function(startTime) {
			var currentConfiguration = _coarseBar.configurations[_coarseBar.configurationIndex],
				segmentCount = Math.ceil(_width / _coarseBar.segmentWidth) + 2,
				timeWindowData = _coarseBar.timeWindow.data()[0],
				halfOfTimeWindow = _width / 2,
				now = Date.now(),
				nowPosition,
				originalTimes = {
					coarseStart: _coarseBar.startTime,
					coarseStop: _coarseBar.stopTime,
					fineStart: _fineBar.start,
					fineStop: _fineBar.stop
				},
				newStartTime = startTime,
				indicatedTime;

			// yes, _coarseRenderTimeline normally does this, but we must pre-calculate it now in order to prevent users from going into the future
			_coarseBar.stopTime = newStartTime + segmentCount * currentConfiguration.milliseconds;
			_coarseBar.scale.domain([new Date(newStartTime), new Date(_coarseBar.stopTime)]);

			// and yes, _fineRenderTimeline normally does this too
			_fineBar.start = _coarseBar.scale.invert(timeWindowData.x).getTime();
			_fineBar.stop = _coarseBar.scale.invert(timeWindowData.x + timeWindowData.width).getTime();
			_fineBar.scale.domain([new Date(_fineBar.start), new Date(_fineBar.stop)]);

			// adjust the start time so the user doesn't go past 50% live on the time window
			nowPosition = _fineBar.scale(now);

			if (halfOfTimeWindow > nowPosition) {
				// calculate the the difference in milliseconds and subtract that from newStartTime
				newStartTime -= (_fineBar.scale.invert(halfOfTimeWindow - nowPosition).getTime() - _fineBar.scale.invert(0).getTime());
			}

			// normalize startTime
			newStartTime = currentConfiguration.rounder(newStartTime);

			// if startTime is the same as newStartTime, roll all the changes above back and don't animate
			if (originalTimes.coarseStart === newStartTime) {
				// roll back
				_coarseBar.startTime = originalTimes.coarseStart;
				_coarseBar.stopTime = originalTimes.coarseStop;
				_coarseBar.scale.domain([new Date(_coarseBar.startTime), new Date(_coarseBar.stopTime)]);
				_fineBar.start = originalTimes.fineStart;
				_fineBar.stop = originalTimes.fineStop;
				_fineBar.scale.domain([new Date(_fineBar.start), new Date(_fineBar.stop)]);
			}
			else {
				_coarseBar.startTime = newStartTime;

				// re-render the coarse bar
				_coarseBarSplit((_coarseBar.startTime < originalTimes.coarseStart) ? 0 : segmentCount, function() {
					_coarseRenderTimeline();

					indicatedTime = _fineBar.scale.invert(_timeIndicator.container.data()[0].x).getTime();

					if (indicatedTime < _fineBar.start || indicatedTime > _fineBar.stop) {
						// the user has dragged the current time out of view, yank it back to the left
						_options.onTimeChange(_fineBar.start, 'via-coarse');
					}
				});

				// re-render the fine bar
				_coarseBar.stopTime = _coarseBar.startTime + segmentCount * currentConfiguration.milliseconds;
				_coarseBar.scale.domain([new Date(_coarseBar.startTime), new Date(_coarseBar.stopTime)]);
				_fineBar.start = _coarseBar.scale.invert(timeWindowData.x).getTime();
				_fineBar.stop = _coarseBar.scale.invert(timeWindowData.x + timeWindowData.width).getTime();

				_fineRenderTimeline(true);

				// if the current time is not within the fine time range, force the needle to the beginning of the range
				if (_fineBar.start > _currentTime && _fineBar.stop < _currentTime) {
					_options.onTimeChange(_fineBar.start, 'via-coarse');
				}
			}
		};

		that.getFineTime = function() {
			return [_fineBar.start, _fineBar.stop];
		};

		that.setFineTime = function(startTime) {
			var fineBarLength = _fineBar.stop - _fineBar.start;

			if (_mode === 'live') {
				_currentTime = startTime;
				_fineBar.start = startTime - _fineBar.liveHalfTime;
				_fineBar.stop = startTime + _fineBar.liveHalfTime;

				// re-render the fine timeline
				_fineRenderTimeline(true);
			}
			else {
				if (_fineBar.stop > startTime && _fineBar.start <= startTime) {
					_currentTime = startTime;

					// re-render the fine timeline
					_fineRenderTimeline(false);
				}
				else {
					// if the requested startTime was outside of the time window, we make some adjustments
					_fineBar.stop = startTime + fineBarLength;
					_fineBar.start = startTime;
					_setupCoarseBarAtTime();

					// re-render the fine timeline
					_fineRenderTimeline(true);
				}
			}
		};

		that.getTimeLengthFromPixels = function(pixelCount) {
			return _fineBar.scale.invert(pixelCount).getTime() - _fineBar.scale.invert(0).getTime();
		};

		that.anchorMotionEventsToTop = function(anchor) {
			_motionEventsAnchored = anchor;

			if (_mode !== 'live') {
				_fineClearEnhancements();
				_fineRenderEnhancements();
			}
		};

		that.setMotionEvents = function(motionEvents) {
			_motionEventData = motionEvents;

			_debounceFineRenderEnhancements();
		};

		that.trackLive = function() {
			var previousMode = _mode,
				diff;

			_mode = 'live';

			// create the fine bar if necessary
			_createFineBar();

			if (_coarseBar.created === true) {
				// if the coarse bar was previously created, we need to hide the left/right drag handles over the fine bar during live mode
				_fineBar.sizerGroup.style('display', 'none');

				setTimeout(function() {
					_coarseBar.container.style('display', 'none');
				}, 250);
			}

			if (previousMode === 'playback') {
				// reset liveHalfTime in case the user adjusted the time window
				_fineBar.liveHalfTime = Math.floor((_fineBar.stop - _fineBar.start) / 2);
			}

			// make some minor adjustments to the fineBar
			_fineBar.scale.range([0, _calculateCurrentWidth()]);

			if (previousMode === 'playback') {
				// make the adjustment based on the xOffset changes above
				diff = _fineBar.scale.invert(_fineBar.xOffset) - _fineBar.scale.invert(0);
				_fineBar.start -= diff;
				_fineBar.stop += diff;
			}

			_fineBar.xOffset = 0;

			_fineBar.axisGroup.attr('transform', 'translate(0, ' + _fineBar.yOffset + ')');
			_fineBar.axisMinorGroup.attr('transform', 'translate(0, ' + _fineBar.yOffset + ')');
			_fineBar.axisMotionGroup.attr('transform', 'translate(0, ' + _fineBar.yOffset + ')');

			// center the time indicator
			_positionFineIndicator(_width / 2);
		};

		that.trackPlayback = function() {
			var previousMode = _mode,
				diff;

			_mode = 'playback';

			if (_coarseBar.created === true) {
				// if the coarse bar was previously created, we need to display the left/right drag handles over the fine bar
				_fineBar.sizerGroup.style('display', '');

				_coarseBar.container.style('display', '');
			}

			// create the fine bar if necessary
			_createFineBar();

			// create the coarse bar if necessary
			_createCoarseBar();

			// make some minor adjustments to the fineBar to account for the left/right drag handles
			_fineBar.xOffset = 25;

			_fineBar.axisGroup.attr('transform', 'translate(' + _fineBar.xOffset + ', ' + _fineBar.yOffset + ')');
			_fineBar.axisMinorGroup.attr('transform', 'translate(' + _fineBar.xOffset + ', ' + _fineBar.yOffset + ')');
			_fineBar.axisMotionGroup.attr('transform', 'translate(' + _fineBar.xOffset + ', ' + _fineBar.yOffset + ')');

			// if the user was just viewing live and transitioned to playback, some initialization needs to occur
			if (previousMode === 'live') {
				// make the adjustment based on the xOffset changes above
				diff = _fineBar.scale.invert(_fineBar.xOffset) - _fineBar.scale.invert(0);
				_fineBar.start += diff;
				_fineBar.stop -= diff;

				_setupCoarseBarAtTime();
			}

			_fineBar.scale.range([0, _calculateCurrentWidth()]);

			// re-render the fine timeline
			_fineRenderTimeline();
		};

		that.resize = function() {
			var ticks,
				fineWidth;

			_width = _container.node().clientWidth;
			fineWidth = _calculateCurrentWidth();
			_xOffset = _container.node().getBoundingClientRect().left;

			// necessary for FireFox
			_svg.attr('width', _width);

			if (_fineBar.created === true) {
				// update the fine bar's scale
				_fineBar.scale.range([0, fineWidth]);

				// determine the appropriate number of ticks depending on the width
				ticks = _calculateTicksForWidth();

				_fineBar.axis
					.ticks(ticks);
			}

			if (_mode === 'live') {
				// in live mode, we just need to center the time indicator... the repeated calls to setTime will re-render the timeline
				_positionFineIndicator(_width / 2);
			}
			else {
				_fineRenderTimeline();

				if (_coarseBar.created === true) {
					_coarseRenderTimeline();
				}
			}

			// always position the right handle if created
			if (_coarseBar.created === true) {
				_fineBar.rightHandleGrabber
					.attr('transform', 'translate(' + (_width - 25) + ')')
					.data([
						{ x: _width - 25 }
					]);
			}
		};

		that.adjustScale = function(step, completeCallback) {
			var segmentCount = Math.ceil(_width / _coarseBar.segmentWidth) + 2,
				atSegment = step === -1 ? segmentCount - 1 : 1,
				splitFunc = step === -1 ? _coarseBarUnsplit : _coarseBarSplit,
				currentConfigurationIndex = _coarseBar.configurationIndex,
				newConfigurationIndex,
				savedTimeWindowPosition = {
					start: _fineBar.start,
					stop: _fineBar.stop
				};

			atSegment = Math.floor((_coarseBar.scale(_fineBar.start)) / _coarseBar.segmentWidth) + 1;

			if (atSegment === 0 || atSegment === 1 || atSegment === 2) {
				atSegment = step === -1 ? 3 : 0;
			}

			// determine what the new _coarseBar.configurationIndex will be
			newConfigurationIndex = Math.min(Math.max(0, _coarseBar.configurationIndex + step), _coarseBar.configurations.length - 1);

			if (currentConfigurationIndex !== newConfigurationIndex) {
				splitFunc(atSegment, function() {
					_adjustScaleComplete(atSegment, newConfigurationIndex, true, savedTimeWindowPosition);
					completeCallback();
				});
			}
			else {
				completeCallback();
			}
		};

		that.destroy = function() {
			_log('timeline destroy');

			// nothing to destroy currently
		};

		// initialize

		_init();
	};
}(console));
