(function($console) {
	'use strict';

	window.ipConfigure = window.ipConfigure || {};
	ipConfigure.components = ipConfigure.components || {};

	/**
	 * Video class, injects an FBGST object into the element indicated by the id parameter.
	 * Exposes events and allows control of the video.
	 *
	 * @param {string} id
	 * @param {object} options backgroundColor (html color), debug (bool), requestPlaybackStream (callback), requestBaseUrl (callback)
	 *
	 * @namespace ipConfigure.components.video
	 */
	ipConfigure.components.video = function(id, options) {
		var that = this,
			_pluginContainer = document.querySelector('#' + id),
			_pauseBetweenStopAndPlay = 10,
			_plugin,
			_pluginValidated = false,
			_startTime,
			_currentTime = 0,
			_pauseOnStart = false,
			_monitorPlaybackTimer,
			_defaultOptions = {
				authentication: {
					username: '',
					password: ''
				},
				backgroundColor: '#EEEEEE',
				debug: false,
				getTime: function() {
					_log('getTime event not subscribed to');
					return 0;
				},
				onAudioKnown: function(hasAudio) {
					_log('onAudioKnown, hasAudio: ', hasAudio);
				},
				onEndReached: function() {
				},
				onPlayFromTime: function() {
				},
				onPlayLive: function() {
				},
				onStateChange: function(s) {
					_log('onStateChange event not subscribed to: ', s);
				},
				stream: undefined
			},
			_options = ipConfigure.core.mergeOptions(_defaultOptions, options),
			_stateLookup = {
				void: 0,
				null: 1,
				ready: 2,
				loading: 3,
				playing: 4
			};

		// private functions

		function _error() {
			$console.error.apply($console, arguments);
		}

		function _log() {
			if (_options.debug === true) {
				$console.log.apply($console, arguments);
			}
		}

		function _addPluginEvent(eventName, callback) {
			if (_plugin.attachEvent) {
				_plugin.attachEvent('on' + eventName, callback);
			}
			else {
				_plugin.addEventListener(eventName, callback, false);
			}
		}

		function _createComponent() {
			var html = '<div class="plugin" style="width: 100%; height: 100%;">' +
				'   <object type="application/x-fbgst" width="100%" height="100%">' +
				'      <param name="windowless" value="true" />' +
				'      <param name="gstScaling" value="true" />' +
				'      <param name="bgcolor" value="' + _options.backgroundColor + '" />' +
				'      <param name="mute" value="true" />' +
				'   </object>' +
				'   <div class="plugin-mask"></div>' +
				'</div>';

			_pluginContainer.className += ' ipc-video';
			_pluginContainer.innerHTML = html;

			_plugin = _pluginContainer.querySelector('object');
		}

		function _validateComponent() {
			if (_plugin.version === undefined) {
				_options.onStateChange('error', 'Unable to play video -- please install FBGST v' + window.requiredFbgstVersion);
				_pluginValidated = false;
			}
			else if (ipConfigure.core.compareVersions(_plugin.version, window.requiredFbgstVersion) === false) {
				_options.onStateChange('error', 'FBGST v' + window.requiredFbgstVersion + ' required -- you are currently using v' + _plugin.version);
				_pluginValidated = false;
			}
			else {
				_pluginValidated = true;
			}
		}

		function _createEventSubscriptions() {
			_addPluginEvent('StateChanged', function(i) {
				switch (i) {
					case _stateLookup.void:
						_options.onStateChange('void');
						break;
					case _stateLookup.null:
						_options.onStateChange('null');
						break;
					case _stateLookup.ready:
						_options.onStateChange('ready');
						break;
					case _stateLookup.loading:
						_options.onStateChange('loading');
						break;
					case _stateLookup.playing:
						if (_pauseOnStart === true) {
							that.stop();
						}
						else {
							// only state change if they aren't pausing on start to avoid any flickering -- stop will trigger another StateChange
							_options.onStateChange('playing');
						}
						break;
				}
			});

			_addPluginEvent('EncounteredError', function(s) {
				_error('EncounteredError', s);
				_options.onStateChange('error', 'Unable to display video at this time');
			});

			_addPluginEvent('StreamEnded', function() {
				_log('StreamEnded');
				_options.onStateChange('error', 'Unable to display video at this time');
			});

			_addPluginEvent('AudioKnown', function(hasAudio) {
				_options.onAudioKnown(hasAudio);
			});
		}

		function _monitorPlayback() {
			if (that.isPlaying === true) {
				// simply calculate the currentTime using plugin.time this has the advantage of automatically taking pause into account
				_currentTime = _startTime + _plugin.time;
			}

			_monitorPlaybackTimer = setTimeout(_monitorPlayback, 100);
		}

		function _pluginStop() {
			try {
				_plugin.stop();
			}
			catch (e) {
				_log('call to _plugin.stop() failed', e);
			}
		}

		function _init() {
			_log('video options', _options);

			// create the FBGST video
			_createComponent();

			// make sure FBGST is installed and up-to-date
			_validateComponent();

			if (_pluginValidated === true) {
				// create subscriptions to necessary events
				_createEventSubscriptions();
			}
		}

		// public properties

		that.isLive = false;

		that.isPlaying = false;

		that.options = _options;

		// public methods

		that.getVolume = function() {
			return _plugin.volume;
		};

		that.setVolume = function(value) {
			_plugin.volume = value;
			_plugin.mute = false;
		};

		that.getCurrentTime = function() {
			if (_pluginValidated === false) {
				return 0;
			}

			return _currentTime;
		};

		that.getVideo = function() {
			return _plugin;
		};

		/**
		 * @param {number} startTime number of milliseconds since 1970
		 * @param {boolean} autoPause automatically pauses the video when it begins playing when true
		 */
		that.playFromTime = function(startTime, autoPause) {
			var serverDifferential = window.clientServerTime && typeof window.clientServerTime.diff === 'number' ? window.clientServerTime.diff : 0;

			if (_pluginValidated === false) {
				return false;
			}

			// set necessary properties
			that.isLive = false;
			that.isPlaying = true;

			// keep track of when we started playing (indicated by the client)
			_startTime = startTime;
			_currentTime = startTime;

			// reset local variables
			_pauseOnStart = autoPause;

			_options.requestPlaybackStream(startTime - serverDifferential, function(src) {
				if (src === null) {
					_options.onStateChange('error', 'unable to obtain playback stream');
				}
				else {
					// set the loading state and stop whatever is playing
					_pluginStop();

					// delay playback to give fbgst a chance to process the stop event
					setTimeout(function() {
						_plugin.play(src, _options.authentication.username, _options.authentication.password);

						// begin monitoring playback
						_monitorPlayback();
					}, _pauseBetweenStopAndPlay);
				}
			});

			// let em know
			_options.onPlayFromTime();

			// clear previously monitored playback
			if (_monitorPlaybackTimer) {
				clearTimeout(_monitorPlaybackTimer);
			}
		};

		that.playLive = function() {
			if (_pluginValidated === false) {
				return false;
			}

			// set necessary properties
			that.isLive = true;
			that.isPlaying = true;

			// keep track of when we started playing so we can use FBGST's .time property later to calculate play time
			_startTime = Date.now();
			_currentTime = _startTime;

			// let em know
			_options.onPlayLive();

			// delay playback to give fbgst a chance to process the stop event
			setTimeout(function() {
				_plugin.play(_options.stream.liveStream, _options.authentication.username, _options.authentication.password);

				// clear previously monitored playback
				if (_monitorPlaybackTimer) {
					clearTimeout(_monitorPlaybackTimer);
				}

				// begin monitoring playback
				_monitorPlayback();
			}, _pauseBetweenStopAndPlay);
		};

		that.stop = function() {
			if (_pluginValidated === false) {
				return false;
			}

			// clean up public properties
			that.isLive = false;
			that.isPlaying = false;

			// clean local bits
			_currentTime = 0;

			// stop playback
			_pluginStop();

			// stop monitoring playback
			if (_monitorPlaybackTimer) {
				clearTimeout(_monitorPlaybackTimer);
			}
		};

		that.destroy = function() {
			_log('video destroy');

			// stop monitoring playback
			if (_monitorPlaybackTimer) {
				clearTimeout(_monitorPlaybackTimer);
			}
		};

		// initialize

		_init();
	};
}(console));
